#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Tests  _symbolic_operator.py."""
import copy
import unittest
import warnings

import numpy
from openfermion.config import EQ_TOLERANCE
from openfermion.utils._testing_utils import EqualsTester

from openfermion.ops._symbolic_operator import SymbolicOperator


class DummyOperator1(SymbolicOperator):
    """Subclass of SymbolicOperator created for testing purposes."""
    actions = (1, 0)
    action_strings = ('^', '')
    action_before_index = False
    different_indices_commute = False


class DummyOperator2(SymbolicOperator):
    """Subclass of SymbolicOperator created for testing purposes."""
    actions = ('X', 'Y', 'Z')
    action_strings = ('X', 'Y', 'Z')
    action_before_index = True
    different_indices_commute = True


class GeneralTest(unittest.TestCase):
    """General tests."""

    def test_init_single_factor(self):
        """Test initialization of the form DummyOperator((index, action))."""
        equals_tester = EqualsTester(self)

        group_1 = [DummyOperator1((3, 0)), DummyOperator1(((3, 0),))]
        group_2 = [DummyOperator2((5, 'X')), DummyOperator2(((5, 'X'),))]
        group_3 = [DummyOperator2((5, 'X'), .5),
                   DummyOperator2(((5, 'X'),), .5)]

        equals_tester.add_equality_group(*group_1)
        equals_tester.add_equality_group(*group_2)
        equals_tester.add_equality_group(*group_3)

    def test_eq_and_neq(self):
        """Test == and !=."""
        equals_tester = EqualsTester(self)

        zeros_1 = [DummyOperator1(),
                   DummyOperator1('1^ 0', 0.),
                   DummyOperator1('1^ 0', -1j) * 0]

        zeros_2 = [DummyOperator2(),
                   DummyOperator2(((1, 'Y'), (0, 'X')), 0.),
                   DummyOperator2(((1, 'Y'), (0, 'X')), -1j) * 0]

        different_ops_1 = [DummyOperator1(((1, 0),), -0.1j),
                           DummyOperator1(((1, 1),), -0.1j),
                           (DummyOperator1(((1, 0),), -0.1j) +
                            DummyOperator1(((1, 1),), -0.1j))]

        different_ops_2 = [DummyOperator2(((1, 'Y'),), -0.1j),
                           DummyOperator2(((1, 'X'),), -0.1j),
                           (DummyOperator2(((1, 'Y'),), -0.1j) +
                            DummyOperator2(((2, 'Y'),), -0.1j))]

        equals_tester.add_equality_group(*zeros_1)
        equals_tester.add_equality_group(*zeros_2)
        for op in different_ops_1:
            equals_tester.add_equality_group(op)
        for op in different_ops_2:
            equals_tester.add_equality_group(op)

    def test_many_body_order(self):
        """Test computing the many-body order."""
        zero = DummyOperator1()
        identity = DummyOperator2(())

        op1 = DummyOperator1('0^ 3 5^ 6')
        op2 = op1 + DummyOperator1('8^ 3')
        op3 = op2 + DummyOperator1('1^ 2 3^ 4 5 ')

        op4 = DummyOperator2('X0 X1 Y3')
        op5 = op4 - DummyOperator2('Z0')
        op6 = op5 - DummyOperator2('Z1 Z2 Y3 Y4 Y9 Y10')
        op7 = op5 - DummyOperator2('Z1 Z2 Y3 Y4 Y9 Y10', EQ_TOLERANCE / 2.)

        self.assertEqual(zero.many_body_order(), 0)
        self.assertEqual(identity.many_body_order(), 0)
        self.assertEqual(op1.many_body_order(), 4)
        self.assertEqual(op2.many_body_order(), 4)
        self.assertEqual(op3.many_body_order(), 5)
        self.assertEqual(op4.many_body_order(), 3)
        self.assertEqual(op5.many_body_order(), 3)
        self.assertEqual(op6.many_body_order(), 6)
        self.assertEqual(op7.many_body_order(), 3)


class SymbolicOperatorTest1(unittest.TestCase):
    """Test the subclass DummyOperator1."""

    def test_init_defaults(self):
        loc_op = DummyOperator1()
        self.assertEqual(len(loc_op.terms), 0)

    def test_init_tuple_real_coefficient(self):
        loc_op = ((0, 1), (5, 0), (6, 1))
        coefficient = 0.5
        fermion_op = DummyOperator1(loc_op, coefficient)
        self.assertEqual(len(fermion_op.terms), 1)
        self.assertEqual(fermion_op.terms[tuple(loc_op)], coefficient)

    def test_init_tuple_complex_coefficient(self):
        loc_op = ((0, 1), (5, 0), (6, 1))
        coefficient = 0.6j
        fermion_op = DummyOperator1(loc_op, coefficient)
        self.assertEqual(len(fermion_op.terms), 1)
        self.assertEqual(fermion_op.terms[tuple(loc_op)], coefficient)

    def test_init_tuple_npfloat64_coefficient(self):
        loc_op = ((0, 1), (5, 0), (6, 1))
        coefficient = numpy.float64(2.303)
        fermion_op = DummyOperator1(loc_op, coefficient)
        self.assertEqual(len(fermion_op.terms), 1)
        self.assertEqual(fermion_op.terms[tuple(loc_op)], coefficient)

    def test_init_tuple_npcomplex128_coefficient(self):
        loc_op = ((0, 1), (5, 0), (6, 1))
        coefficient = numpy.complex128(-1.123j + 43.7)
        fermion_op = DummyOperator1(loc_op, coefficient)
        self.assertEqual(len(fermion_op.terms), 1)
        self.assertEqual(fermion_op.terms[tuple(loc_op)], coefficient)

    def test_init_list_real_coefficient(self):
        loc_op = [(0, 1), (5, 0), (6, 1)]
        coefficient = 1. / 3
        fermion_op = DummyOperator1(loc_op, coefficient)
        self.assertEqual(len(fermion_op.terms), 1)
        self.assertEqual(fermion_op.terms[tuple(loc_op)], coefficient)

    def test_init_list_complex_coefficient(self):
        loc_op = [(0, 1), (5, 0), (6, 1)]
        coefficient = 2j / 3.
        fermion_op = DummyOperator1(loc_op, coefficient)
        self.assertEqual(len(fermion_op.terms), 1)
        self.assertEqual(fermion_op.terms[tuple(loc_op)], coefficient)

    def test_init_list_npfloat64_coefficient(self):
        loc_op = [(0, 1), (5, 0), (6, 1)]
        coefficient = numpy.float64(2.3037)
        fermion_op = DummyOperator1(loc_op, coefficient)
        self.assertEqual(len(fermion_op.terms), 1)
        self.assertEqual(fermion_op.terms[tuple(loc_op)], coefficient)

    def test_init_list_npcomplex128_coefficient(self):
        loc_op = [(0, 1), (5, 0), (6, 1)]
        coefficient = numpy.complex128(-1.1237j + 43.37)
        fermion_op = DummyOperator1(loc_op, coefficient)
        self.assertEqual(len(fermion_op.terms), 1)
        self.assertEqual(fermion_op.terms[tuple(loc_op)], coefficient)

    def test_identity_is_multiplicative_identity(self):
        u = DummyOperator1.identity()
        f = DummyOperator1(((0, 1), (5, 0), (6, 1)), 0.6j)
        g = DummyOperator1(((0, 0), (5, 0), (6, 1)), 0.3j)
        h = f + g
        self.assertTrue(f == u * f)
        self.assertTrue(f == f * u)
        self.assertTrue(g == u * g)
        self.assertTrue(g == g * u)
        self.assertTrue(h == u * h)
        self.assertTrue(h == h * u)

        u *= h
        self.assertTrue(h == u)
        self.assertFalse(f == u)

        # Method always returns new instances.
        self.assertFalse(DummyOperator1.identity() == u)

    def test_zero_is_additive_identity(self):
        o = DummyOperator1.zero()
        f = DummyOperator1(((0, 1), (5, 0), (6, 1)), 0.6j)
        g = DummyOperator1(((0, 0), (5, 0), (6, 1)), 0.3j)
        h = f + g
        self.assertTrue(f == o + f)
        self.assertTrue(f == f + o)
        self.assertTrue(g == o + g)
        self.assertTrue(g == g + o)
        self.assertTrue(h == o + h)
        self.assertTrue(h == h + o)

        o += h
        self.assertTrue(h == o)
        self.assertFalse(f == o)

        # Method always returns new instances.
        self.assertFalse(DummyOperator1.zero() == o)

    def test_zero_is_multiplicative_nil(self):
        o = DummyOperator1.zero()
        u = DummyOperator1.identity()
        f = DummyOperator1(((0, 1), (5, 0), (6, 1)), 0.6j)
        g = DummyOperator1(((0, 0), (5, 0), (6, 1)), 0.3j)
        self.assertTrue(o == o * u)
        self.assertTrue(o == o * f)
        self.assertTrue(o == o * g)
        self.assertTrue(o == o * (f + g))

    def test_init_str(self):
        fermion_op = DummyOperator1('0^ 5 12^', -1.)
        correct = ((0, 1), (5, 0), (12, 1))
        self.assertIn(correct, fermion_op.terms)
        self.assertEqual(fermion_op.terms[correct], -1.0)

    def test_init_long_str(self):
        fermion_op = DummyOperator1(
                '(-2.0+3.0j) [0^ 1] +\n\n -1.0[ 2^ 3 ] - []', -1.)
        correct = \
            DummyOperator1('0^ 1', complex(2., -3.)) + \
            DummyOperator1('2^ 3', 1.) + \
            DummyOperator1('', 1.)
        self.assertEqual(len((fermion_op-correct).terms), 0)
        reparsed_op = DummyOperator1(str(fermion_op))
        self.assertEqual(len((fermion_op-reparsed_op).terms), 0)

        fermion_op = DummyOperator1('1.7 [3^ 2] - 8 [4^]')
        correct = DummyOperator1('3^ 2', 1.7) + DummyOperator1('4^', -8.)
        self.assertEqual(len((fermion_op-correct).terms), 0)

        fermion_op = DummyOperator1('-(2.3 + 1.7j) [3^ 2]')
        correct = DummyOperator1('3^ 2', complex(-2.3, -1.7))
        self.assertEqual(len((fermion_op-correct).terms), 0)

    def test_merges_multiple_whitespace(self):
        fermion_op = DummyOperator1('        \n ')
        self.assertEqual(fermion_op.terms, {(): 1})

    def test_init_str_identity(self):
        fermion_op = DummyOperator1('')
        self.assertIn((), fermion_op.terms)

    def test_init_bad_term(self):
        with self.assertRaises(ValueError):
            DummyOperator1(2)

    def test_init_bad_coefficient(self):
        with self.assertRaises(ValueError):
            DummyOperator1('0^', "0.5")

    def test_init_bad_action_str(self):
        with self.assertRaises(ValueError):
            DummyOperator1('0-')

    def test_init_bad_action_tuple(self):
        with self.assertRaises(ValueError):
            DummyOperator1(((0, 2),))

    def test_init_bad_tuple(self):
        with self.assertRaises(ValueError):
            DummyOperator1(((0, 1, 1),))

    def test_init_bad_str(self):
        with self.assertRaises(ValueError):
            DummyOperator1('^')

    def test_init_bad_mode_num(self):
        with self.assertRaises(ValueError):
            DummyOperator1('-1^')

    def test_init_invalid_tensor_factor(self):
        with self.assertRaises(ValueError):
            DummyOperator1(((-2, 1), (1, 0)))

    def test_DummyOperator1(self):
        op = DummyOperator1((), 3.)
        self.assertTrue(op == DummyOperator1(()) * 3.)

    def test_imul_inplace(self):
        fermion_op = DummyOperator1("1^")
        prev_id = id(fermion_op)
        fermion_op *= 3.
        self.assertEqual(id(fermion_op), prev_id)
        self.assertEqual(fermion_op.terms[((1, 1),)], 3.)

    def test_imul_scalar_real(self):
        loc_op = ((1, 0), (2, 1))
        multiplier = 0.5
        fermion_op = DummyOperator1(loc_op)
        fermion_op *= multiplier
        self.assertEqual(fermion_op.terms[loc_op], multiplier)

    def test_imul_scalar_complex(self):
        loc_op = ((1, 0), (2, 1))
        multiplier = 0.6j
        fermion_op = DummyOperator1(loc_op)
        fermion_op *= multiplier
        self.assertEqual(fermion_op.terms[loc_op], multiplier)

    def test_imul_scalar_npfloat64(self):
        loc_op = ((1, 0), (2, 1))
        multiplier = numpy.float64(2.303)
        fermion_op = DummyOperator1(loc_op)
        fermion_op *= multiplier
        self.assertEqual(fermion_op.terms[loc_op], multiplier)

    def test_imul_scalar_npcomplex128(self):
        loc_op = ((1, 0), (2, 1))
        multiplier = numpy.complex128(-1.123j + 1.7911)
        fermion_op = DummyOperator1(loc_op)
        fermion_op *= multiplier
        self.assertEqual(fermion_op.terms[loc_op], multiplier)

    def test_imul_fermion_op(self):
        op1 = DummyOperator1(((0, 1), (3, 0), (8, 1), (8, 0), (11, 1)), 3.j)
        op2 = DummyOperator1(((1, 1), (3, 1), (8, 0)), 0.5)
        op1 *= op2
        correct_coefficient = 1.j * 3.0j * 0.5
        correct_term = ((0, 1), (3, 0), (8, 1), (8, 0), (11, 1),
                        (1, 1), (3, 1), (8, 0))
        self.assertEqual(len(op1.terms), 1)
        self.assertIn(correct_term, op1.terms)

    def test_imul_fermion_op_2(self):
        op3 = DummyOperator1(((1, 1), (0, 0)), -1j)
        op4 = DummyOperator1(((1, 0), (0, 1), (2, 1)), -1.5)
        op3 *= op4
        op4 *= op3
        self.assertIn(((1, 1), (0, 0), (1, 0), (0, 1), (2, 1)), op3.terms)
        self.assertEqual(op3.terms[((1, 1), (0, 0), (1, 0), (0, 1), (2, 1))],
                         1.5j)

    def test_imul_fermion_op_duplicate_term(self):
        op1 = DummyOperator1('1 2 3')
        op1 += DummyOperator1('1 2')
        op1 += DummyOperator1('1')

        op2 = DummyOperator1('3')
        op2 += DummyOperator1('2 3')

        op1 *= op2
        self.assertAlmostEqual(op1.terms[((1, 0), (2, 0), (3, 0))], 2.)

    def test_imul_bidir(self):
        op_a = DummyOperator1(((1, 1), (0, 0)), -1j)
        op_b = DummyOperator1(((1, 1), (0, 1), (2, 1)), -1.5)
        op_a *= op_b
        op_b *= op_a
        self.assertIn(((1, 1), (0, 0), (1, 1), (0, 1), (2, 1)), op_a.terms)
        self.assertEqual(op_a.terms[((1, 1), (0, 0), (1, 1), (0, 1), (2, 1))],
                         1.5j)
        self.assertIn(((1, 1), (0, 1), (2, 1),
                       (1, 1), (0, 0), (1, 1), (0, 1), (2, 1)), op_b.terms)
        self.assertEqual(op_b.terms[((1, 1), (0, 1), (2, 1),
                                     (1, 1), (0, 0),
                                     (1, 1), (0, 1), (2, 1))], -2.25j)

    def test_imul_bad_multiplier(self):
        op = DummyOperator1(((1, 1), (0, 1)), -1j)
        with self.assertRaises(TypeError):
            op *= "1"

    def test_mul_by_scalarzero(self):
        op = DummyOperator1(((1, 1), (0, 1)), -1j) * 0
        self.assertNotIn(((0, 1), (1, 1)), op.terms)
        self.assertIn(((1, 1), (0, 1)), op.terms)
        self.assertEqual(op.terms[((1, 1), (0, 1))], 0.0)

    def test_mul_bad_multiplier(self):
        op = DummyOperator1(((1, 1), (0, 1)), -1j)
        with self.assertRaises(TypeError):
            op = op * "0.5"

    def test_mul_out_of_place(self):
        op1 = DummyOperator1(((0, 1), (3, 1), (3, 0), (11, 1)), 3.j)
        op2 = DummyOperator1(((1, 1), (3, 1), (8, 0)), 0.5)
        op3 = op1 * op2
        correct_coefficient = 3.0j * 0.5
        correct_term = ((0, 1), (3, 1), (3, 0), (11, 1),
                        (1, 1), (3, 1), (8, 0))
        self.assertTrue(op1 == DummyOperator1(
            ((0, 1), (3, 1), (3, 0), (11, 1)), 3.j))
        self.assertTrue(op2 == DummyOperator1(((1, 1), (3, 1), (8, 0)), 0.5))
        self.assertTrue(op3 == DummyOperator1(correct_term,
                                              correct_coefficient))

    def test_mul_npfloat64(self):
        op = DummyOperator1(((1, 0), (3, 1)), 0.5)
        res = op * numpy.float64(0.5)
        self.assertTrue(res == DummyOperator1(((1, 0), (3, 1)), 0.5 * 0.5))

    def test_mul_multiple_terms(self):
        op = DummyOperator1(((1, 0), (8, 1)), 0.5)
        op += DummyOperator1(((1, 1), (9, 1)), 1.4j)
        res = op * op
        correct = DummyOperator1(((1, 0), (8, 1), (1, 0), (8, 1)), 0.5 ** 2)
        correct += (DummyOperator1(((1, 0), (8, 1), (1, 1), (9, 1)), 0.7j) +
                    DummyOperator1(((1, 1), (9, 1), (1, 0), (8, 1)), 0.7j))
        correct += DummyOperator1(((1, 1), (9, 1), (1, 1), (9, 1)), 1.4j ** 2)
        self.assertTrue(res == correct)

    def test_rmul_scalar_real(self):
        op = DummyOperator1(((1, 1), (3, 0), (8, 1)), 0.5)
        multiplier = 0.5
        res1 = op * multiplier
        res2 = multiplier * op
        self.assertTrue(res1 == res2)

    def test_rmul_scalar_complex(self):
        op = DummyOperator1(((1, 1), (3, 0), (8, 1)), 0.5)
        multiplier = 0.6j
        res1 = op * multiplier
        res2 = multiplier * op
        self.assertTrue(res1 == res2)

    def test_rmul_scalar_npfloat64(self):
        op = DummyOperator1(((1, 1), (3, 0), (8, 1)), 0.5)
        multiplier = numpy.float64(2.303)
        res1 = op * multiplier
        res2 = multiplier * op
        self.assertTrue(res1 == res2)

    def test_rmul_scalar_npcomplex128(self):
        op = DummyOperator1(((1, 1), (3, 0), (8, 1)), 0.5)
        multiplier = numpy.complex128(-1.5j + 7.7)
        res1 = op * multiplier
        res2 = multiplier * op
        self.assertTrue(res1 == res2)

    def test_rmul_bad_multiplier(self):
        op = DummyOperator1(((1, 1), (3, 0), (8, 1)), 0.5)
        with self.assertRaises(TypeError):
            op = "0.5" * op

    def test_truediv_and_div_real(self):
        op = DummyOperator1(((1, 1), (3, 0), (8, 1)), 0.5)
        divisor = 0.5
        original = copy.deepcopy(op)
        res = op / divisor
        correct = op * (1. / divisor)
        self.assertTrue(res == correct)
        # Test if done out of place
        self.assertTrue(op == original)

    def test_truediv_and_div_complex(self):
        op = DummyOperator1(((1, 1), (3, 0), (8, 1)), 0.5)
        divisor = 0.6j
        original = copy.deepcopy(op)
        res = op / divisor
        correct = op * (1. / divisor)
        self.assertTrue(res == correct)
        # Test if done out of place
        self.assertTrue(op == original)

    def test_truediv_and_div_npfloat64(self):
        op = DummyOperator1(((1, 1), (3, 0), (8, 1)), 0.5)
        divisor = numpy.float64(2.303)
        original = copy.deepcopy(op)
        res = op / divisor
        correct = op * (1. / divisor)
        self.assertTrue(res == correct)
        # Test if done out of place
        self.assertTrue(op == original)

    def test_truediv_and_div_npcomplex128(self):
        op = DummyOperator1(((1, 1), (3, 0), (8, 1)), 0.5)
        divisor = numpy.complex128(566.4j + 0.3)
        original = copy.deepcopy(op)
        res = op / divisor
        correct = op * (1. / divisor)
        self.assertTrue(res == correct)
        # Test if done out of place
        self.assertTrue(op == original)

    def test_truediv_bad_divisor(self):
        op = DummyOperator1(((1, 1), (3, 0), (8, 1)), 0.5)
        with self.assertRaises(TypeError):
            op = op / "0.5"

    def test_itruediv_and_idiv_real(self):
        op = DummyOperator1(((1, 1), (3, 0), (8, 1)), 0.5)
        divisor = 0.5
        original = copy.deepcopy(op)
        correct = op * (1. / divisor)
        op /= divisor
        self.assertTrue(op == correct)
        # Test if done in-place
        self.assertFalse(op == original)

    def test_itruediv_and_idiv_complex(self):
        op = DummyOperator1(((1, 1), (3, 0), (8, 1)), 0.5)
        divisor = 0.6j
        original = copy.deepcopy(op)
        correct = op * (1. / divisor)
        op /= divisor
        self.assertTrue(op == correct)
        # Test if done in-place
        self.assertFalse(op == original)

    def test_itruediv_and_idiv_npfloat64(self):
        op = DummyOperator1(((1, 1), (3, 0), (8, 1)), 0.5)
        divisor = numpy.float64(2.3030)
        original = copy.deepcopy(op)
        correct = op * (1. / divisor)
        op /= divisor
        self.assertTrue(op == correct)
        # Test if done in-place
        self.assertFalse(op == original)

    def test_itruediv_and_idiv_npcomplex128(self):
        op = DummyOperator1(((1, 1), (3, 0), (8, 1)), 0.5)
        divisor = numpy.complex128(12.3 + 7.4j)
        original = copy.deepcopy(op)
        correct = op * (1. / divisor)
        op /= divisor
        self.assertTrue(op == correct)
        # Test if done in-place
        self.assertFalse(op == original)

    def test_itruediv_bad_divisor(self):
        op = DummyOperator1(((1, 1), (3, 0), (8, 1)), 0.5)
        with self.assertRaises(TypeError):
            op /= "0.5"

    def test_iadd_different_term(self):
        term_a = ((1, 1), (3, 0), (8, 1))
        term_b = ((1, 1), (3, 1), (8, 0))
        a = DummyOperator1(term_a, 1.0)
        a += DummyOperator1(term_b, 0.5)
        self.assertEqual(len(a.terms), 2)
        self.assertEqual(a.terms[term_a], 1.0)
        self.assertEqual(a.terms[term_b], 0.5)
        a += DummyOperator1(term_b, 0.5)
        self.assertEqual(len(a.terms), 2)
        self.assertEqual(a.terms[term_a], 1.0)
        self.assertEqual(a.terms[term_b], 1.0)

    def test_iadd_bad_addend(self):
        op = DummyOperator1((), 1.0)
        with self.assertRaises(TypeError):
            op += "0.5"

    def test_add(self):
        term_a = ((1, 1), (3, 0), (8, 1))
        term_b = ((1, 0), (3, 0), (8, 1))
        a = DummyOperator1(term_a, 1.0)
        b = DummyOperator1(term_b, 0.5)
        res = a + b + b
        self.assertEqual(len(res.terms), 2)
        self.assertEqual(res.terms[term_a], 1.0)
        self.assertEqual(res.terms[term_b], 1.0)
        # Test out of place
        self.assertTrue(a == DummyOperator1(term_a, 1.0))
        self.assertTrue(b == DummyOperator1(term_b, 0.5))

    def test_add_bad_addend(self):
        op = DummyOperator1((), 1.0)
        with self.assertRaises(TypeError):
            _ = op + "0.5"

    def test_sub(self):
        term_a = ((1, 1), (3, 1), (8, 1))
        term_b = ((1, 0), (3, 1), (8, 1))
        a = DummyOperator1(term_a, 1.0)
        b = DummyOperator1(term_b, 0.5)
        res = a - b
        self.assertEqual(len(res.terms), 2)
        self.assertEqual(res.terms[term_a], 1.0)
        self.assertEqual(res.terms[term_b], -0.5)
        res2 = b - a
        self.assertEqual(len(res2.terms), 2)
        self.assertEqual(res2.terms[term_a], -1.0)
        self.assertEqual(res2.terms[term_b], 0.5)

    def test_sub_bad_subtrahend(self):
        op = DummyOperator1((), 1.0)
        with self.assertRaises(TypeError):
            _ = op - "0.5"

    def test_isub_different_term(self):
        term_a = ((1, 1), (3, 1), (8, 0))
        term_b = ((1, 0), (3, 1), (8, 1))
        a = DummyOperator1(term_a, 1.0)
        a -= DummyOperator1(term_b, 0.5)
        self.assertEqual(len(a.terms), 2)
        self.assertEqual(a.terms[term_a], 1.0)
        self.assertEqual(a.terms[term_b], -0.5)
        a -= DummyOperator1(term_b, 0.5)
        self.assertEqual(len(a.terms), 2)
        self.assertEqual(a.terms[term_a], 1.0)
        self.assertEqual(a.terms[term_b], -1.0)

    def test_isub_bad_addend(self):
        op = DummyOperator1((), 1.0)
        with self.assertRaises(TypeError):
            op -= "0.5"

    def test_neg(self):
        op = DummyOperator1(((1, 1), (3, 1), (8, 1)), 0.5)
        _ = -op
        # out of place
        self.assertTrue(op == DummyOperator1(((1, 1), (3, 1), (8, 1)), 0.5))
        correct = -1.0 * op
        self.assertTrue(correct == -op)

    def test_pow_square_term(self):
        coeff = 6.7j
        ops = ((3, 1), (1, 0), (4, 1))
        term = DummyOperator1(ops, coeff)
        squared = term ** 2
        expected = DummyOperator1(ops + ops, coeff ** 2)
        self.assertTrue(squared == term * term)
        self.assertTrue(squared == expected)

    def test_pow_zero_term(self):
        coeff = 6.7j
        ops = ((3, 1), (1, 0), (4, 1))
        term = DummyOperator1(ops, coeff)
        zerod = term ** 0
        expected = DummyOperator1(())
        self.assertTrue(expected == zerod)

    def test_pow_one_term(self):
        coeff = 6.7j
        ops = ((3, 1), (1, 0), (4, 1))
        term = DummyOperator1(ops, coeff)
        self.assertTrue(term == term ** 1)

    def test_pow_high_term(self):
        coeff = 6.7j
        ops = ((3, 1), (1, 0), (4, 1))
        term = DummyOperator1(ops, coeff)
        high = term ** 10
        expected = DummyOperator1(ops * 10, coeff ** 10)
        self.assertTrue(expected == high)

    def test_pow_neg_error(self):
        with self.assertRaises(ValueError):
            DummyOperator1() ** -1

    def test_pow_nonint_error(self):
        with self.assertRaises(ValueError):
            DummyOperator1('3 2^') ** 0.5

    def test_compress_terms(self):
        op = (DummyOperator1('3^ 1', 0.3 + 3e-11j) +
              DummyOperator1('2^ 3', 5e-10) +
              DummyOperator1('1^ 3', 1e-3))
        op_compressed = (DummyOperator1('3^ 1', 0.3) +
                         DummyOperator1('1^ 3', 1e-3))
        op.compress(1e-7)
        self.assertTrue(op_compressed == op)

    def test_str(self):
        op = DummyOperator1(((1, 1), (3, 0), (8, 1)), 0.5)
        self.assertEqual(str(op), "0.5 [1^ 3 8^]")
        op2 = DummyOperator1((), 2)
        self.assertEqual(str(op2), "2 []")
        op3 = DummyOperator1()
        self.assertEqual(str(op3), "0")

    def test_rep(self):
        op = DummyOperator1(((1, 1), (3, 0), (8, 1)), 0.5)
        # Not necessary, repr could do something in addition
        self.assertEqual(repr(op), str(op))


class SymbolicOperatorTest2(unittest.TestCase):
    """Test the subclass DummyOperator2."""
    def test_init_defaults(self):
        loc_op = DummyOperator2()
        self.assertTrue(len(loc_op.terms) == 0)

    def test_init_tuple(self):
        coefficient = 0.5
        loc_op = ((0, 'X'), (5, 'Y'), (6, 'Z'))
        qubit_op = DummyOperator2(loc_op, coefficient)
        self.assertTrue(len(qubit_op.terms) == 1)
        self.assertTrue(qubit_op.terms[loc_op] == coefficient)

    def test_init_list(self):
        coefficient = 0.6j
        loc_op = [(0, 'X'), (5, 'Y'), (6, 'Z')]
        qubit_op = DummyOperator2(loc_op, coefficient)
        self.assertTrue(len(qubit_op.terms) == 1)
        self.assertTrue(qubit_op.terms[tuple(loc_op)] == coefficient)

    def test_init_str(self):
        qubit_op = DummyOperator2('X0 Y5 Z12', -1.)
        correct = ((0, 'X'), (5, 'Y'), (12, 'Z'))
        self.assertTrue(correct in qubit_op.terms)
        self.assertTrue(qubit_op.terms[correct] == -1.0)

    def test_init_long_str(self):
        qubit_op = DummyOperator2(
                '(-2.0+3.0j) [X0 Y1] +\n\n -1.0[ X2 Y3 ] - []', -1.)
        correct = \
            DummyOperator2('X0 Y1', complex(2., -3.)) + \
            DummyOperator2('X2 Y3', 1.) + \
            DummyOperator2('', 1.)
        self.assertEqual(len((qubit_op-correct).terms), 0)
        reparsed_op = DummyOperator2(str(qubit_op))
        self.assertEqual(len((qubit_op-reparsed_op).terms), 0)

        qubit_op = DummyOperator2('[X0 X1] + [Y0 Y1]')
        correct = DummyOperator2('X0 X1') + DummyOperator2('Y0 Y1')
        self.assertTrue(qubit_op == correct)
        self.assertTrue(qubit_op == DummyOperator2(str(qubit_op)))

    def test_init_str_identity(self):
        qubit_op = DummyOperator2('', 2.)
        self.assertTrue(len(qubit_op.terms) == 1)
        self.assertTrue(() in qubit_op.terms)
        self.assertAlmostEqual(qubit_op.terms[()], 2.)

    def test_init_bad_term(self):
        with self.assertRaises(ValueError):
            qubit_op = DummyOperator2(2)

    def test_init_bad_coefficient(self):
        with self.assertRaises(ValueError):
            qubit_op = DummyOperator2('X0', "0.5")

    def test_init_bad_action(self):
        with self.assertRaises(ValueError):
            qubit_op = DummyOperator2('Q0')

    def test_init_bad_action_in_tuple(self):
        with self.assertRaises(ValueError):
            qubit_op = DummyOperator2(((1, 'Q'),))

    def test_init_bad_qubit_num_in_tuple(self):
        with self.assertRaises(ValueError):
            qubit_op = DummyOperator2((("1", 'X'),))

    def test_init_bad_tuple(self):
        with self.assertRaises(ValueError):
            qubit_op = DummyOperator2(((0, 1, 'X'),))

    def test_init_bad_str(self):
        with self.assertRaises(ValueError):
            qubit_op = DummyOperator2('X')

    def test_init_bad_qubit_num(self):
        with self.assertRaises(ValueError):
            qubit_op = DummyOperator2('X-1')

    def test_compress(self):
        a = DummyOperator2('X0', .9e-12)
        self.assertTrue(len(a.terms) == 1)
        a.compress()
        self.assertTrue(len(a.terms) == 0)
        a = DummyOperator2('X0', 1. + 1j)
        a.compress(.5)
        self.assertTrue(len(a.terms) == 1)
        for term in a.terms:
            self.assertTrue(a.terms[term] == 1. + 1j)
        a = DummyOperator2('X0', 1.1 + 1j)
        a.compress(1.)
        self.assertTrue(len(a.terms) == 1)
        for term in a.terms:
            self.assertTrue(a.terms[term] == 1.1)
        a = DummyOperator2('X0', 1.1 + 1j) + DummyOperator2('X1', 1.e-6j)
        a.compress()
        self.assertTrue(len(a.terms) == 2)
        for term in a.terms:
            self.assertTrue(isinstance(a.terms[term], complex))
        a.compress(1.e-5)
        self.assertTrue(len(a.terms) == 1)
        for term in a.terms:
            self.assertTrue(isinstance(a.terms[term], complex))
        a.compress(1.)
        self.assertTrue(len(a.terms) == 1)
        for term in a.terms:
            self.assertTrue(isinstance(a.terms[term], float))

    def test_rmul_scalar(self):
        multiplier = 0.5
        op = DummyOperator2(((1, 'X'), (3, 'Y'), (8, 'Z')), 0.5)
        res1 = op * multiplier
        res2 = multiplier * op
        self.assertTrue(res1 == res2)

    def test_rmul_bad_multiplier(self):
        op = DummyOperator2(((1, 'X'), (3, 'Y'), (8, 'Z')), 0.5)
        with self.assertRaises(TypeError):
            op = "0.5" * op

    def test_truediv_and_div(self):
        divisor = 0.6j
        op = DummyOperator2(((1, 'X'), (3, 'Y'), (8, 'Z')), 0.5)
        op2 = copy.deepcopy(op)
        original = copy.deepcopy(op)
        res = op / divisor
        res2 = op2.__div__(divisor)  # To test python 2 version as well
        correct = op * (1. / divisor)
        self.assertTrue(res == correct)
        self.assertTrue(res2 == correct)
        # Test if done out of place
        self.assertTrue(op == original)
        self.assertTrue(op2 == original)

    def test_truediv_bad_divisor(self):
        op = DummyOperator2(((1, 'X'), (3, 'Y'), (8, 'Z')), 0.5)
        with self.assertRaises(TypeError):
            op = op / "0.5"

    def test_itruediv_and_idiv(self):
        divisor = 2
        op = DummyOperator2(((1, 'X'), (3, 'Y'), (8, 'Z')), 0.5)
        op2 = copy.deepcopy(op)
        original = copy.deepcopy(op)
        correct = op * (1. / divisor)
        op /= divisor
        op2.__idiv__(divisor)  # To test python 2 version as well
        self.assertTrue(op == correct)
        self.assertTrue(op2 == correct)
        # Test if done in-place
        self.assertTrue(not op == original)
        self.assertTrue(not op2 == original)

    def test_itruediv_bad_divisor(self):
        op = DummyOperator2(((1, 'X'), (3, 'Y'), (8, 'Z')), 0.5)
        with self.assertRaises(TypeError):
            op /= "0.5"

    def test_iadd_cancellation(self):
        term_a = ((1, 'X'), (3, 'Y'), (8, 'Z'))
        term_b = ((1, 'X'), (3, 'Y'), (8, 'Z'))
        a = DummyOperator2(term_a, 1.0)
        a += DummyOperator2(term_b, -1.0)
        self.assertTrue(len(a.terms) == 0)

    def test_iadd_different_term(self):
        term_a = ((1, 'X'), (3, 'Y'), (8, 'Z'))
        term_b = ((1, 'Z'), (3, 'Y'), (8, 'Z'))
        a = DummyOperator2(term_a, 1.0)
        a += DummyOperator2(term_b, 0.5)
        self.assertTrue(len(a.terms) == 2)
        self.assertAlmostEqual(a.terms[term_a], 1.0)
        self.assertAlmostEqual(a.terms[term_b], 0.5)
        a += DummyOperator2(term_b, 0.5)
        self.assertTrue(len(a.terms) == 2)
        self.assertAlmostEqual(a.terms[term_a], 1.0)
        self.assertAlmostEqual(a.terms[term_b], 1.0)

    def test_iadd_bad_addend(self):
        op = DummyOperator2((), 1.0)
        with self.assertRaises(TypeError):
            op += "0.5"

    def test_add(self):
        term_a = ((1, 'X'), (3, 'Y'), (8, 'Z'))
        term_b = ((1, 'Z'), (3, 'Y'), (8, 'Z'))
        a = DummyOperator2(term_a, 1.0)
        b = DummyOperator2(term_b, 0.5)
        res = a + b + b
        self.assertTrue(len(res.terms) == 2)
        self.assertAlmostEqual(res.terms[term_a], 1.0)
        self.assertAlmostEqual(res.terms[term_b], 1.0)
        # Test out of place
        self.assertTrue(a == DummyOperator2(term_a, 1.0))
        self.assertTrue(b == DummyOperator2(term_b, 0.5))

    def test_add_bad_addend(self):
        op = DummyOperator2((), 1.0)
        with self.assertRaises(TypeError):
            op = op + "0.5"

    def test_sub(self):
        term_a = ((1, 'X'), (3, 'Y'), (8, 'Z'))
        term_b = ((1, 'Z'), (3, 'Y'), (8, 'Z'))
        a = DummyOperator2(term_a, 1.0)
        b = DummyOperator2(term_b, 0.5)
        res = a - b
        self.assertTrue(len(res.terms) == 2)
        self.assertAlmostEqual(res.terms[term_a], 1.0)
        self.assertAlmostEqual(res.terms[term_b], -0.5)
        res2 = b - a
        self.assertTrue(len(res2.terms) == 2)
        self.assertAlmostEqual(res2.terms[term_a], -1.0)
        self.assertAlmostEqual(res2.terms[term_b], 0.5)

    def test_sub_bad_subtrahend(self):
        op = DummyOperator2((), 1.0)
        with self.assertRaises(TypeError):
            op = op - "0.5"

    def test_isub_different_term(self):
        term_a = ((1, 'X'), (3, 'Y'), (8, 'Z'))
        term_b = ((1, 'Z'), (3, 'Y'), (8, 'Z'))
        a = DummyOperator2(term_a, 1.0)
        a -= DummyOperator2(term_b, 0.5)
        self.assertTrue(len(a.terms) == 2)
        self.assertAlmostEqual(a.terms[term_a], 1.0)
        self.assertAlmostEqual(a.terms[term_b], -0.5)
        a -= DummyOperator2(term_b, 0.5)
        self.assertTrue(len(a.terms) == 2)
        self.assertAlmostEqual(a.terms[term_a], 1.0)
        self.assertAlmostEqual(a.terms[term_b], -1.0)

    def test_isub_bad_addend(self):
        op = DummyOperator2((), 1.0)
        with self.assertRaises(TypeError):
            op -= "0.5"

    def test_neg(self):
        op = DummyOperator2(((1, 'X'), (3, 'Y'), (8, 'Z')), 0.5)
        -op
        # out of place
        self.assertTrue(op == DummyOperator2(((1, 'X'), (3, 'Y'), (8, 'Z')),
                                             0.5))
        correct = -1.0 * op
        self.assertTrue(correct == -op)

    def test_str(self):
        op = DummyOperator2(((1, 'X'), (3, 'Y'), (8, 'Z')), 0.5)
        self.assertEqual(str(op), "0.5 [X1 Y3 Z8]")
        op2 = DummyOperator2((), 2)
        self.assertEqual(str(op2), "2 []")

    def test_str_empty(self):
        op = DummyOperator2()
        self.assertEqual(str(op), '0')

    def test_str_out_of_order(self):
        op = DummyOperator2(((3, 'Y'), (1, 'X'), (8, 'Z')), 0.5)
        self.assertEqual(str(op), '0.5 [X1 Y3 Z8]')

    def test_str_multiple_terms(self):
        op = DummyOperator2(((1, 'X'), (3, 'Y'), (8, 'Z')), 0.5)
        op += DummyOperator2(((1, 'Y'), (3, 'Y'), (8, 'Z')), 0.6)
        self.assertTrue((str(op) == "0.5 [X1 Y3 Z8] +\n0.6 [Y1 Y3 Z8]" or
                         str(op) == "0.6 [Y1 Y3 Z8] +\n0.5 [X1 Y3 Z8]"))
        op2 = DummyOperator2((), 2)
        self.assertEqual(str(op2), "2 []")

    def test_rep(self):
        op = DummyOperator2(((1, 'X'), (3, 'Y'), (8, 'Z')), 0.5)
        # Not necessary, repr could do something in addition
        self.assertEqual(repr(op), str(op))

    def test_norm(self):
        op = DummyOperator2(((1, 'X'), (3, 'Y'), (8, 'Z')), 1)
        op += DummyOperator2(((2, 'Z'), (3, 'Y')), 1)
        self.assertAlmostEqual(op.induced_norm(2), numpy.sqrt(2.))

    def test_tracenorm_zero(self):
        op = SymbolicOperator()
        self.assertFalse(op.induced_norm())


class DeprecatedFunctionsTest(unittest.TestCase):
    """Tests for deprecated functions."""

    def test_warnings(self):
        """Test that warnings are raised appropriately."""
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")

            op1 = DummyOperator1()
            op1.isclose(op1)

            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[0].category, DeprecationWarning))
            self.assertTrue("deprecated" in str(w[0].message))

    def test_isclose_zero_terms_1(self):
        op = DummyOperator1('1^ 0', -1j) * 0
        self.assertTrue(op == DummyOperator1())
        self.assertTrue(DummyOperator1() == op)

    def test_isclose_different_terms_1(self):
        a = DummyOperator1(((1, 0),), -0.1j)
        b = DummyOperator1(((1, 1),), -0.1j)
        self.assertFalse(a == b)

    def test_isclose_different_num_terms_1(self):
        a = DummyOperator1(((1, 0),), -0.1j)
        a += DummyOperator1(((1, 1),), -0.1j)
        b = DummyOperator1(((1, 0),), -0.1j)
        self.assertFalse(b == a)
        self.assertFalse(a == b)

    def test_isclose_zero_terms_2(self):
        op = DummyOperator2(((1, 'Y'), (0, 'X')), -1j) * 0
        self.assertTrue(op == DummyOperator2())
        self.assertTrue(DummyOperator2((), 0.0) == op)

    def test_isclose_different_terms_2(self):
        a = DummyOperator2(((1, 'Y'),), -0.1j)
        b = DummyOperator2(((1, 'X'),), -0.1j)
        self.assertTrue(not a == b)
        self.assertTrue(not b == a)

    def test_isclose_different_num_terms_2(self):
        a = DummyOperator2(((1, 'Y'),), -0.1j)
        a += DummyOperator2(((2, 'Y'),), -0.1j)
        b = DummyOperator2(((1, 'X'),), -0.1j)
        self.assertTrue(not b == a)
        self.assertTrue(not a == b)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""This module contains functions for compiling circuits to prepare
Slater determinants and fermionic Gaussian states."""
from __future__ import absolute_import

import numpy

from openfermion.config import EQ_TOLERANCE
from openfermion.ops import QuadraticHamiltonian
from openfermion.ops._quadratic_hamiltonian import swap_columns


def gaussian_state_preparation_circuit(
        quadratic_hamiltonian, occupied_orbitals=None):
    """Obtain the description of a circuit which prepares a fermionic Gaussian
    state.

    Fermionic Gaussian states can be regarded as eigenstates of quadratic
    Hamiltonians. If the Hamiltonian conserves particle number, then these are
    just Slater determinants. See arXiv:1711.05395 for a detailed description
    of how this procedure works.

    The circuit description is returned as a sequence of elementary
    operations; operations that can be performed in parallel are grouped
    together. Each elementary operation is either

    - the string 'pht', indicating the particle-hole transformation
      on the last fermionic mode, which is the operator :math:`\mathcal{B}`
      such that

      .. math::

          \\begin{align}
              \mathcal{B} a_N \mathcal{B}^\dagger &= a_N^\dagger,\\\\
              \mathcal{B} a_j \mathcal{B}^\dagger &= a_j, \\quad
                  j = 1, \ldots, N-1,
          \end{align}

      or

    - a tuple :math:`(i, j, \\theta, \\varphi)`, indicating the operation

      .. math::
          \exp[i \\varphi a_j^\dagger a_j]
          \exp[\\theta (a_i^\dagger a_j - a_j^\dagger a_i)],

      a Givens rotation of modes :math:`i` and :math:`j` by angles
      :math:`\\theta` and :math:`\\varphi`.

    Args:
        quadratic_hamiltonian(QuadraticHamiltonian):
            The Hamiltonian whose eigenstate is desired.
        occupied_orbitals(list):
            A list of integers representing the indices of the occupied
            orbitals in the desired Gaussian state. If this is None
            (the default), then it is assumed that the ground state is
            desired, i.e., the orbitals with negative energies are filled.

    Returns
    -------
        circuit_description (list[tuple]):
            A list of operations describing the circuit. Each operation
            is a tuple of objects describing elementary operations that
            can be performed in parallel. Each elementary operation
            is either the string 'pht', indicating a particle-hole
            transformation on the last fermionic mode, or a tuple of
            the form :math:`(i, j, \\theta, \\varphi)`,
            indicating a Givens rotation
            of modes :math:`i` and :math:`j` by angles :math:`\\theta`
            and :math:`\\varphi`.
        start_orbitals (list):
            The occupied orbitals to start with. This describes the
            initial state that the circuit should be applied to: it should
            be a Slater determinant (in the computational basis) with these
            orbitals filled.
    """
    if not isinstance(quadratic_hamiltonian, QuadraticHamiltonian):
        raise ValueError('Input must be an instance of QuadraticHamiltonian.')

    if quadratic_hamiltonian.conserves_particle_number:
        # The Hamiltonian conserves particle number, so we don't need
        # to use the most general procedure.
        hermitian_matrix = quadratic_hamiltonian.combined_hermitian_part
        energies, diagonalizing_unitary = numpy.linalg.eigh(hermitian_matrix)

        if occupied_orbitals is None:
            # The ground state is desired, so we fill the orbitals that have
            # negative energy
            num_negative_energies = numpy.count_nonzero(
                energies < -EQ_TOLERANCE)
            occupied_orbitals = range(num_negative_energies)

        # Get the unitary rows which represent the Slater determinant
        slater_determinant_matrix = diagonalizing_unitary.T[occupied_orbitals]

        # Get the circuit description
        circuit_description = slater_determinant_preparation_circuit(
            slater_determinant_matrix)
        start_orbitals = range(len(occupied_orbitals))
    else:
        # The Hamiltonian does not conserve particle number, so we
        # need to use the most general procedure.
        diagonalizing_unitary = (
            quadratic_hamiltonian.diagonalizing_bogoliubov_transform())

        # Get the unitary rows which represent the Gaussian unitary
        gaussian_unitary_matrix = diagonalizing_unitary[
            quadratic_hamiltonian.n_qubits:]

        # Get the circuit description
        decomposition, left_decomposition, diagonal, left_diagonal = (
            fermionic_gaussian_decomposition(gaussian_unitary_matrix))
        if occupied_orbitals is None:
            # The ground state is desired, so the circuit should be applied
            # to the vaccuum state
            start_orbitals = []
            circuit_description = list(reversed(decomposition))
        else:
            start_orbitals = occupied_orbitals
            # The circuit won't be applied to the ground state, so we need to
            # use left_decomposition
            circuit_description = list(reversed(
                decomposition + left_decomposition))

    return circuit_description, start_orbitals


def slater_determinant_preparation_circuit(slater_determinant_matrix):
    """Obtain the description of a circuit which prepares a Slater determinant.

    The input is an :math:`N_f \\times N` matrix :math:`Q` with orthonormal
    rows. Such a matrix describes the Slater determinant

    .. math::

        b^\dagger_1 \cdots b^\dagger_{N_f} \lvert \\text{vac} \\rangle,

    where

    .. math::

        b^\dagger_j = \sum_{k = 1}^N Q_{jk} a^\dagger_k.

    The output is the description of a circuit which prepares this
    Slater determinant, up to a global phase.
    The starting state which the circuit should be applied to
    is a Slater determinant (in the computational basis) with
    the first :math:`N_f` orbitals filled.

    Args:
        slater_determinant_matrix: The matrix :math:`Q` which describes the
            Slater determinant to be prepared.
    Returns:
        circuit_description:
            A list of operations describing the circuit. Each operation
            is a tuple of elementary operations that can be performed in
            parallel. Each elementary operation is a tuple of the form
            :math:`(i, j, \\theta, \\varphi)`, indicating a Givens rotation
            of modes :math:`i` and :math:`j` by angles :math:`\\theta`
            and :math:`\\varphi`.
    """
    decomposition, left_unitary, diagonal = givens_decomposition(
        slater_determinant_matrix)
    circuit_description = list(reversed(decomposition))
    return circuit_description


def fermionic_gaussian_decomposition(unitary_rows):
    """Decompose a matrix into a sequence of Givens rotations and
    particle-hole transformations on the last fermionic mode.

    The input is an :math:`N \\times 2N` matrix :math:`W` with orthonormal
    rows. Furthermore, :math:`W` must have the block form

    .. math::

        W = ( W_1 \hspace{4pt} W_2 )

    where :math:`W_1` and :math:`W_2` satisfy

    .. math::

        W_1  W_1^\dagger + W_2  W_2^\dagger &= I

        W_1  W_2^T + W_2  W_1^T &= 0.

    Then :math:`W` can be decomposed as

    .. math::

        V  W  U^\dagger = ( 0 \hspace{6pt} D )

    where :math:`V` and :math:`U` are unitary matrices and :math:`D`
    is a diagonal unitary matrix. Furthermore, :math:`U` can be decomposed
    as follows:

    .. math::

        U = B G_{k} \cdots B G_3 G_2 B G_1 B,

    where each :math:`G_i` is a Givens rotation, and :math:`B` represents
    swapping the :math:`N`-th column with the :math:`2N`-th column,
    which corresponds to a particle-hole transformation
    on the last fermionic mode. This particle-hole transformation maps
    :math:`a^\dagger_N` to :math:`a_N` and vice versa, while leaving the
    other fermionic ladder operators invariant.

    The decomposition of :math:`U` is returned as a list of tuples of objects
    describing rotations and particle-hole transformations. The list looks
    something like [('pht', ), (G_1, ), ('pht', G_2), ... ].
    The objects within a tuple are either the string 'pht', which indicates
    a particle-hole transformation on the last fermionic mode, or a tuple
    of the form :math:`(i, j, \\theta, \\varphi)`, which indicates a
    Givens rotation of rows :math:`i` and :math:`j` by angles
    :math:`\\theta` and :math:`\\varphi`.

    The matrix :math:`V^T D^*` can also be decomposed as a sequence of
    Givens rotations. This decomposition is needed for a circuit that
    prepares an excited state.

    Args:
        unitary_rows(ndarray): A matrix with orthonormal rows and
            additional structure described above.

    Returns
    -------
        decomposition (list[tuple]):
            The decomposition of :math:`U`.
        left_decomposition (list[tuple]):
            The decomposition of :math:`V^T D^*`.
        diagonal (ndarray):
            A list of the nonzero entries of :math:`D`.
        left_diagonal (ndarray):
            A list of the nonzero entries left from the decomposition
            of :math:`V^T D^*`.
    """
    current_matrix = numpy.copy(unitary_rows)
    n, p = current_matrix.shape

    # Check that p = 2 * n
    if p != 2 * n:
        raise ValueError('The input matrix must have twice as many columns '
                         'as rows.')

    # Check that left and right parts of unitary_rows satisfy the constraints
    # necessary for the transformed fermionic operators to satisfy
    # the fermionic anticommutation relations
    left_part = unitary_rows[:, :n]
    right_part = unitary_rows[:, n:]
    constraint_matrix_1 = (left_part.dot(left_part.T.conj()) +
                           right_part.dot(right_part.T.conj()))
    constraint_matrix_2 = (left_part.dot(right_part.T) +
                           right_part.dot(left_part.T))
    discrepancy_1 = numpy.amax(abs(constraint_matrix_1 - numpy.eye(n)))
    discrepancy_2 = numpy.amax(abs(constraint_matrix_2))
    if discrepancy_1 > EQ_TOLERANCE or discrepancy_2 > EQ_TOLERANCE:
        raise ValueError('The input matrix does not satisfy the constraints '
                         'necessary for a proper transformation of the '
                         'fermionic ladder operators.')

    # Compute left_unitary using Givens rotations
    left_unitary = numpy.eye(n, dtype=complex)
    for k in range(n - 1):
        # Zero out entries in column k
        for l in range(n - 1 - k):
            # Zero out entry in row l if needed
            if abs(current_matrix[l, k]) > EQ_TOLERANCE:
                givens_rotation = givens_matrix_elements(
                    current_matrix[l, k], current_matrix[l + 1, k])
                # Apply Givens rotation
                givens_rotate(current_matrix, givens_rotation, l, l + 1)
                givens_rotate(left_unitary, givens_rotation, l, l + 1)

    # Initialize list to store decomposition of current_matrix
    decomposition = []
    # There are 2 * n - 1 iterations (that is the circuit depth)
    for k in range(2 * n - 1):
        # Initialize the list of parallel operations to perform
        # in this iteration
        parallel_ops = []

        # Perform a particle-hole transformation if necessary
        if k % 2 == 0 and abs(current_matrix[k // 2, n - 1]) > EQ_TOLERANCE:
            parallel_ops.append('pht')
            swap_columns(current_matrix, n - 1, 2 * n - 1)

        # Get the (row, column) indices of elements to zero out in parallel.
        if k < n:
            end_row = k
            end_column = n - 1 - k
        else:
            end_row = n - 1
            end_column = k - (n - 1)
        column_indices = range(end_column, n - 1, 2)
        row_indices = range(end_row, end_row - len(column_indices), -1)
        indices_to_zero_out = zip(row_indices, column_indices)

        for i, j in indices_to_zero_out:
            # Compute the Givens rotation to zero out the (i, j) element,
            # if needed
            left_element = current_matrix[i, j].conj()
            if abs(left_element) > EQ_TOLERANCE:
                # We actually need to perform a Givens rotation
                right_element = current_matrix[i, j + 1].conj()
                givens_rotation = givens_matrix_elements(left_element,
                                                         right_element)

                # Add the parameters to the list
                theta = numpy.arcsin(numpy.real(givens_rotation[1, 0]))
                phi = numpy.angle(givens_rotation[1, 1])
                parallel_ops.append((j, j + 1, theta, phi))

                # Update the matrix
                double_givens_rotate(current_matrix, givens_rotation,
                                     j, j + 1, which='col')

        # If the current list of parallel operations is not empty,
        # append it to the list,
        if parallel_ops:
            decomposition.append(tuple(parallel_ops))

    # Get the diagonal entries
    diagonal = current_matrix[range(n), range(n, 2 * n)]

    # Compute the decomposition of left_unitary^T * diagonal^*
    current_matrix = left_unitary.T
    for k in range(n):
        current_matrix[:, k] *= diagonal[k].conj()
    left_decomposition = []

    for k in range(2 * (n - 1) - 1):
        # Initialize the list of parallel operations to perform
        # in this iteration
        parallel_ops = []

        # Get the (row, column) indices of elements to zero out in parallel.
        if k < n - 1:
            start_row = 0
            start_column = n - 1 - k
        else:
            start_row = k - (n - 2)
            start_column = k - (n - 3)
        column_indices = range(start_column, n, 2)
        row_indices = range(start_row, start_row + len(column_indices))
        indices_to_zero_out = zip(row_indices, column_indices)

        for i, j in indices_to_zero_out:
            # Compute the Givens rotation to zero out the (i, j) element,
            # if needed
            right_element = current_matrix[i, j].conj()
            if abs(right_element) > EQ_TOLERANCE:
                # We actually need to perform a Givens rotation
                left_element = current_matrix[i, j - 1].conj()
                givens_rotation = givens_matrix_elements(left_element,
                                                         right_element,
                                                         which='right')

                # Add the parameters to the list
                theta = numpy.arcsin(numpy.real(givens_rotation[1, 0]))
                phi = numpy.angle(givens_rotation[1, 1])
                parallel_ops.append((j - 1, j, theta, phi))

                # Update the matrix
                givens_rotate(current_matrix, givens_rotation,
                              j - 1, j, which='col')

        # If the current list of parallel operations is not empty,
        # append it to the list,
        if parallel_ops:
            left_decomposition.append(tuple(parallel_ops))

    # Get the diagonal entries
    left_diagonal = current_matrix[range(n), range(n)]

    return decomposition, left_decomposition, diagonal, left_diagonal


def givens_decomposition(unitary_rows):
    """Decompose a matrix into a sequence of Givens rotations.

    The input is an :math:`m \\times n` matrix :math:`Q` with :math:`m \leq n`.
    The rows of :math:`Q` are orthonormal.
    :math:`Q` can be decomposed as follows:

    .. math::

        V Q U^\dagger = D

    where :math:`V` and :math:`U` are unitary matrices, and :math:`D`
    is an :math:`m \\times n` matrix with the
    first :math:`m` columns forming a diagonal matrix and the rest of the
    columns being zero. Furthermore, we can decompose :math:`U` as

    .. math::

        U = G_k ... G_1

    where :math:`G_1, \\ldots, G_k` are complex Givens rotations.
    A Givens rotation is a rotation within the two-dimensional subspace
    spanned by two coordinate axes. Within the two relevant coordinate
    axes, a Givens rotation has the form

    .. math::

        \\begin{pmatrix}
            \\cos(\\theta) & -e^{i \\varphi} \\sin(\\theta) \\\\
            \\sin(\\theta) &     e^{i \\varphi} \\cos(\\theta)
        \\end{pmatrix}.

    Args:
        unitary_rows: A numpy array or matrix with orthonormal rows,
            representing the matrix Q.

    Returns
    -------
        givens_rotations (list[tuple]):
            A list of tuples of objects describing Givens
            rotations. The list looks like [(G_1, ), (G_2, G_3), ... ].
            The Givens rotations within a tuple can be implemented in parallel.
            The description of a Givens rotation is itself a tuple of the
            form :math:`(i, j, \\theta, \\varphi)`, which represents a
            Givens rotation of coordinates
            :math:`i` and :math:`j` by angles :math:`\\theta` and
            :math:`\\varphi`.
        left_unitary (ndarray):
            An :math:`m \\times m` numpy array representing the matrix
            :math:`V`.
        diagonal (ndarray):
            A list of the nonzero entries of :math:`D`.
    """
    current_matrix = numpy.copy(unitary_rows)
    m, n = current_matrix.shape

    # Check that m <= n
    if m > n:
        raise ValueError('The input m x n matrix must have m <= n')

    # Compute left_unitary using Givens rotations
    left_unitary = numpy.eye(m, dtype=complex)
    for k in reversed(range(n - m + 1, n)):
        # Zero out entries in column k
        for l in range(m - n + k):
            # Zero out entry in row l if needed
            if abs(current_matrix[l, k]) > EQ_TOLERANCE:
                givens_rotation = givens_matrix_elements(
                    current_matrix[l, k], current_matrix[l + 1, k])
                # Apply Givens rotation
                givens_rotate(current_matrix, givens_rotation, l, l + 1)
                givens_rotate(left_unitary, givens_rotation, l, l + 1)

    # Compute the decomposition of current_matrix into Givens rotations
    givens_rotations = []
    # If m = n (the matrix is square) then we don't need to perform any
    # Givens rotations!
    if m != n:
        # Get the maximum number of simultaneous rotations that
        # will be performed
        max_simul_rotations = min(m, n - m)
        # There are n - 1 iterations (the circuit depth is n - 1)
        for k in range(n - 1):
            # Get the (row, column) indices of elements to zero out in
            # parallel.
            if k < max_simul_rotations - 1:
                # There are k + 1 elements to zero out
                start_row = 0
                end_row = k + 1
                start_column = n - m - k
                end_column = start_column + 2 * (k + 1)
            elif k > n - 1 - max_simul_rotations:
                # There are n - 1 - k elements to zero out
                start_row = m - (n - 1 - k)
                end_row = m
                start_column = m - (n - 1 - k) + 1
                end_column = start_column + 2 * (n - 1 - k)
            else:
                # There are max_simul_rotations elements to zero out
                if max_simul_rotations == m:
                    start_row = 0
                    end_row = m
                    start_column = n - m - k
                    end_column = start_column + 2 * m
                else:
                    start_row = k + 1 - max_simul_rotations
                    end_row = k + 1
                    start_column = k + 1 - max_simul_rotations + 1
                    end_column = start_column + 2 * max_simul_rotations

            row_indices = range(start_row, end_row)
            column_indices = range(start_column, end_column, 2)
            indices_to_zero_out = zip(row_indices, column_indices)

            parallel_rotations = []
            for i, j in indices_to_zero_out:
                # Compute the Givens rotation to zero out the (i, j) element,
                # if needed
                right_element = current_matrix[i, j].conj()
                if abs(right_element) > EQ_TOLERANCE:
                    # We actually need to perform a Givens rotation
                    left_element = current_matrix[i, j - 1].conj()
                    givens_rotation = givens_matrix_elements(
                        left_element, right_element, which='right')

                    # Add the parameters to the list
                    theta = numpy.arcsin(numpy.real(givens_rotation[1, 0]))
                    phi = numpy.angle(givens_rotation[1, 1])
                    parallel_rotations.append((j - 1, j, theta, phi))

                    # Update the matrix
                    givens_rotate(current_matrix, givens_rotation,
                                  j - 1, j, which='col')

            # If the current list of parallel operations is not empty,
            # append it to the list,
            if parallel_rotations:
                givens_rotations.append(tuple(parallel_rotations))

    # Get the diagonal entries
    diagonal = current_matrix.diagonal()

    return givens_rotations, left_unitary, diagonal


def givens_matrix_elements(a, b, which='left'):
    """Compute the matrix elements of the Givens rotation that zeroes out one
    of two row entries.

    If `which='left'` then returns a matrix G such that

        G * [a  b]^T= [0  r]^T

    otherwise, returns a matrix G such that

        G * [a  b]^T= [r  0]^T

    where r is a complex number.

    Args:
        a(complex or float): A complex number representing the upper row entry
        b(complex or float): A complex number representing the lower row entry
        which(string): Either 'left' or 'right', indicating whether to
            zero out the left element (first argument) or right element
            (second argument). Default is `left`.
    Returns:
        G(ndarray): A 2 x 2 numpy array representing the matrix G.
            The numbers in the first column of G are real.
    """
    # Handle case that a is zero
    if abs(a) < EQ_TOLERANCE:
        cosine = 1.
        sine = 0.
        phase = 1.
    # Handle case that b is zero and a is nonzero
    elif abs(b) < EQ_TOLERANCE:
        cosine = 0.
        sine = 1.
        phase = 1.
    # Handle case that a and b are both nonzero
    else:
        denominator = numpy.sqrt(abs(a) ** 2 + abs(b) ** 2)
        cosine = abs(b) / denominator
        sine = abs(a) / denominator
        sign_b = b / abs(b)
        sign_a = a / abs(a)
        phase = sign_a * sign_b.conjugate()
        # If phase is a real number, convert it to a float
        if numpy.isreal(phase):
            phase = numpy.real(phase)

    # Construct matrix and return
    if which == 'left':
        # We want to zero out a
        if numpy.isreal(a) and numpy.isreal(b):
            givens_rotation = numpy.array([[cosine, -phase * sine],
                                           [phase * sine, cosine]])
        else:
            givens_rotation = numpy.array([[cosine, -phase * sine],
                                           [sine, phase * cosine]])
    elif which == 'right':
        # We want to zero out b
        if numpy.isreal(a) and numpy.isreal(b):
            givens_rotation = numpy.array([[sine, phase * cosine],
                                           [-phase * cosine, sine]])
        else:
            givens_rotation = numpy.array([[sine, phase * cosine],
                                           [cosine, -phase * sine]])
    else:
        raise ValueError('"which" must be equal to "left" or "right".')
    return givens_rotation


def givens_rotate(operator, givens_rotation, i, j, which='row'):
    """Apply a Givens rotation to coordinates i and j of an operator."""
    if which == 'row':
        # Rotate rows i and j
        row_i = operator[i].copy()
        row_j = operator[j].copy()
        operator[i] = (givens_rotation[0, 0] * row_i +
                       givens_rotation[0, 1] * row_j)
        operator[j] = (givens_rotation[1, 0] * row_i +
                       givens_rotation[1, 1] * row_j)
    elif which == 'col':
        # Rotate columns i and j
        col_i = operator[:, i].copy()
        col_j = operator[:, j].copy()
        operator[:, i] = (givens_rotation[0, 0] * col_i +
                          givens_rotation[0, 1].conj() * col_j)
        operator[:, j] = (givens_rotation[1, 0] * col_i +
                          givens_rotation[1, 1].conj() * col_j)
    else:
        raise ValueError('"which" must be equal to "row" or "col".')


def double_givens_rotate(operator, givens_rotation, i, j, which='row'):
    """Apply a double Givens rotation.

    Applies a Givens rotation to coordinates i and j and the the conjugate
    Givens rotation to coordinates n + i and n + j, where
    n = dim(operator) / 2. dim(operator) must be even.
    """
    m, p = operator.shape

    if which == 'row':
        if m % 2 != 0:
            raise ValueError('To apply a double Givens rotation on rows, '
                             'the number of rows must be even.')
        n = m // 2
        # Rotate rows i and j
        givens_rotate(operator[:n], givens_rotation, i, j, which='row')
        # Rotate rows n + i and n + j
        givens_rotate(operator[n:], givens_rotation.conj(), i, j, which='row')
    elif which == 'col':
        if p % 2 != 0:
            raise ValueError('To apply a double Givens rotation on columns, '
                             'the number of columns must be even.')
        n = p // 2
        # Rotate columns i and j
        givens_rotate(operator[:, :n], givens_rotation, i, j, which='col')
        # Rotate cols n + i and n + j
        givens_rotate(operator[:, n:], givens_rotation.conj(), i, j,
                      which='col')
    else:
        raise ValueError('"which" must be equal to "row" or "col".')



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Tests for polynomial_tensor.py."""
from __future__ import absolute_import, division

import unittest

import copy
import numpy

from openfermion.ops import PolynomialTensor
from openfermion.utils._slater_determinants_test import (
    random_quadratic_hamiltonian)


class PolynomialTensorTest(unittest.TestCase):

    def setUp(self):
        self.n_qubits = 2
        self.constant = 23.0

        one_body_a = numpy.zeros((self.n_qubits, self.n_qubits))
        two_body_a = numpy.zeros((self.n_qubits, self.n_qubits,
                                  self.n_qubits, self.n_qubits))
        one_body_a[0, 1] = 2
        one_body_a[1, 0] = 3
        two_body_a[0, 1, 0, 1] = 4
        two_body_a[1, 1, 0, 0] = 5

        self.polynomial_tensor_a = PolynomialTensor(
            {(): self.constant, (1, 0): one_body_a, (1, 1, 0, 0): two_body_a})

        one_body_na = numpy.zeros((self.n_qubits, self.n_qubits))
        two_body_na = numpy.zeros((self.n_qubits, self.n_qubits,
                                   self.n_qubits, self.n_qubits))
        one_body_na[0, 1] = -2
        one_body_na[1, 0] = -3
        two_body_na[0, 1, 0, 1] = -4
        two_body_na[1, 1, 0, 0] = -5
        self.polynomial_tensor_na = PolynomialTensor(
            {(): -self.constant, (1, 0): one_body_na,
             (1, 1, 0, 0): two_body_na})

        one_body_b = numpy.zeros((self.n_qubits, self.n_qubits))
        two_body_b = numpy.zeros((self.n_qubits, self.n_qubits,
                                  self.n_qubits, self.n_qubits))
        one_body_b[0, 1] = 1
        one_body_b[1, 0] = 2
        two_body_b[0, 1, 0, 1] = 3
        two_body_b[1, 0, 0, 1] = 4
        self.polynomial_tensor_b = PolynomialTensor(
            {(): self.constant, (1, 0): one_body_b,
             (1, 1, 0, 0): two_body_b})

        one_body_ab = numpy.zeros((self.n_qubits, self.n_qubits))
        two_body_ab = numpy.zeros((self.n_qubits, self.n_qubits,
                                   self.n_qubits, self.n_qubits))
        one_body_ab[0, 1] = 3
        one_body_ab[1, 0] = 5
        two_body_ab[0, 1, 0, 1] = 7
        two_body_ab[1, 0, 0, 1] = 4
        two_body_ab[1, 1, 0, 0] = 5
        self.polynomial_tensor_ab = PolynomialTensor(
            {(): 2.0 * self.constant, (1, 0): one_body_ab,
             (1, 1, 0, 0): two_body_ab})

        constant_axb = self.constant * self.constant
        one_body_axb = numpy.zeros((self.n_qubits, self.n_qubits))
        two_body_axb = numpy.zeros((self.n_qubits, self.n_qubits,
                                    self.n_qubits, self.n_qubits))
        one_body_axb[0, 1] = 2
        one_body_axb[1, 0] = 6
        two_body_axb[0, 1, 0, 1] = 12
        self.polynomial_tensor_axb = PolynomialTensor(
            {(): constant_axb, (1, 0): one_body_axb,
             (1, 1, 0, 0): two_body_axb})

        self.n_qubits_plus_one = self.n_qubits + 1
        one_body_c = numpy.zeros((self.n_qubits_plus_one,
                                  self.n_qubits_plus_one))
        two_body_c = numpy.zeros((self.n_qubits_plus_one,
                                  self.n_qubits_plus_one,
                                  self.n_qubits_plus_one,
                                  self.n_qubits_plus_one))
        one_body_c[0, 1] = 1
        one_body_c[1, 0] = 2
        two_body_c[0, 1, 0, 1] = 3
        two_body_c[1, 0, 0, 1] = 4
        self.polynomial_tensor_c = PolynomialTensor(
            {(): self.constant, (1, 0): one_body_c,
             (1, 1, 0, 0): two_body_c})

        one_body_hole = numpy.zeros((self.n_qubits, self.n_qubits))
        two_body_hole = numpy.zeros((self.n_qubits, self.n_qubits,
                                     self.n_qubits, self.n_qubits))
        one_body_hole[0, 1] = 2
        one_body_hole[1, 0] = 3
        two_body_hole[0, 1, 0, 1] = 4
        two_body_hole[1, 1, 0, 0] = 5

        self.polynomial_tensor_hole = PolynomialTensor(
            {(): self.constant, (0, 1): one_body_hole,
             (0, 0, 1, 1): two_body_hole})

        one_body_spinful = numpy.zeros((2 * self.n_qubits, 2 * self.n_qubits))
        two_body_spinful = numpy.zeros((2 * self.n_qubits, 2 * self.n_qubits,
                                        2 * self.n_qubits, 2 * self.n_qubits))
        one_body_spinful[0, 1] = 2
        one_body_spinful[1, 0] = 3
        one_body_spinful[2, 3] = 6
        one_body_spinful[3, 2] = 7
        two_body_spinful[0, 1, 0, 1] = 4
        two_body_spinful[1, 1, 0, 0] = 5
        two_body_spinful[2, 1, 2, 3] = 8
        two_body_spinful[3, 3, 2, 2] = 9

        self.polynomial_tensor_spinful = PolynomialTensor(
            {(): self.constant, (1, 0): one_body_spinful,
             (1, 1, 0, 0): two_body_spinful})

    def test_setitem_1body(self):
        expected_one_body_tensor = numpy.array([[0, 3], [2, 0]])
        self.polynomial_tensor_a[(0, 1), (1, 0)] = 3
        self.polynomial_tensor_a[(1, 1), (0, 0)] = 2
        self.assertTrue(numpy.allclose(
            self.polynomial_tensor_a.n_body_tensors[(1, 0)],
            expected_one_body_tensor))

    def test_getitem_1body(self):
        self.assertEqual(self.polynomial_tensor_c[(0, 1), (1, 0)], 1)
        self.assertEqual(self.polynomial_tensor_c[(1, 1), (0, 0)], 2)

    def test_setitem_2body(self):
        self.polynomial_tensor_a[(0, 1), (1, 1), (1, 0), (0, 0)] = 3
        self.polynomial_tensor_a[(1, 1), (0, 1), (0, 0), (1, 0)] = 2
        self.assertEqual(
            self.polynomial_tensor_a.n_body_tensors[
                (1, 1, 0, 0)][0, 1, 1, 0], 3)
        self.assertEqual(
            self.polynomial_tensor_a.n_body_tensors[
                (1, 1, 0, 0)][1, 0, 0, 1], 2)

    def test_getitem_2body(self):
        self.assertEqual(
            self.polynomial_tensor_c[(0, 1), (1, 1), (0, 0), (1, 0)], 3)
        self.assertEqual(
            self.polynomial_tensor_c[(1, 1), (0, 1), (0, 0), (1, 0)], 4)

    def test_invalid_getitem_indexing(self):
        with self.assertRaises(KeyError):
            self.polynomial_tensor_a[(0, 1), (1, 1), (0, 0)]

    def test_invalid_setitem_indexing(self):
        test_tensor = copy.deepcopy(self.polynomial_tensor_a)
        with self.assertRaises(KeyError):
            test_tensor[(0, 1), (1, 1), (0, 0)] = 5

    def test_eq(self):
        self.assertEqual(self.polynomial_tensor_a,
                         self.polynomial_tensor_a)
        self.assertNotEqual(self.polynomial_tensor_a,
                            self.polynomial_tensor_hole)
        self.assertNotEqual(self.polynomial_tensor_a,
                            self.polynomial_tensor_spinful)

    def test_neq(self):
        self.assertNotEqual(self.polynomial_tensor_a,
                            self.polynomial_tensor_b)

    def test_add(self):
        new_tensor = self.polynomial_tensor_a + self.polynomial_tensor_b
        self.assertEqual(new_tensor, self.polynomial_tensor_ab)

    def test_iadd(self):
        new_tensor = copy.deepcopy(self.polynomial_tensor_a)
        new_tensor += self.polynomial_tensor_b
        self.assertEqual(new_tensor, self.polynomial_tensor_ab)

    def test_invalid_addend(self):
        with self.assertRaises(TypeError):
            self.polynomial_tensor_a + 2

    def test_invalid_tensor_shape_add(self):
        with self.assertRaises(TypeError):
            self.polynomial_tensor_a + self.polynomial_tensor_c

    def test_invalid_tensor_keys_add(self):
        with self.assertRaises(TypeError):
            self.polynomial_tensor_a + self.polynomial_tensor_hole

    def test_neg(self):
        self.assertEqual(-self.polynomial_tensor_a,
                         self.polynomial_tensor_na)

    def test_sub(self):
        new_tensor = self.polynomial_tensor_ab - self.polynomial_tensor_b
        self.assertEqual(new_tensor, self.polynomial_tensor_a)

    def test_isub(self):
        new_tensor = copy.deepcopy(self.polynomial_tensor_ab)
        new_tensor -= self.polynomial_tensor_b
        self.assertEqual(new_tensor, self.polynomial_tensor_a)

    def test_invalid_subtrahend(self):
        with self.assertRaises(TypeError):
            self.polynomial_tensor_a - 2

    def test_invalid_tensor_shape_sub(self):
        with self.assertRaises(TypeError):
            self.polynomial_tensor_a - self.polynomial_tensor_c

    def test_invalid_tensor_keys_sub(self):
        with self.assertRaises(TypeError):
            self.polynomial_tensor_a - self.polynomial_tensor_hole

    def test_mul(self):
        new_tensor = self.polynomial_tensor_a * self.polynomial_tensor_b
        self.assertEqual(new_tensor, self.polynomial_tensor_axb)

    def test_imul(self):
        new_tensor = copy.deepcopy(self.polynomial_tensor_a)
        new_tensor *= self.polynomial_tensor_b
        self.assertEqual(new_tensor, self.polynomial_tensor_axb)

    def test_invalid_multiplier(self):
        with self.assertRaises(TypeError):
            self.polynomial_tensor_a * 2

    def test_invalid_tensor_shape_mult(self):
        with self.assertRaises(TypeError):
            self.polynomial_tensor_a * self.polynomial_tensor_c

    def test_invalid_tensor_keys_mult(self):
        with self.assertRaises(TypeError):
            self.polynomial_tensor_a * self.polynomial_tensor_hole

    def test_iter_and_str(self):
        one_body = numpy.zeros((self.n_qubits, self.n_qubits))
        two_body = numpy.zeros((self.n_qubits, self.n_qubits,
                                self.n_qubits, self.n_qubits))
        one_body[0, 1] = 11.0
        two_body[0, 1, 1, 0] = 22.0
        polynomial_tensor = PolynomialTensor(
            {(): self.constant, (1, 0): one_body, (1, 1, 0, 0): two_body})
        want_str = ('() 23.0\n((0, 1), (1, 0)) 11.0\n'
                    '((0, 1), (1, 1), (1, 0), (0, 0)) 22.0\n')
        self.assertEqual(str(polynomial_tensor), want_str)
        self.assertEqual(polynomial_tensor.__repr__(), want_str)

    def test_rotate_basis_identical(self):
        rotation_matrix_identical = numpy.zeros((self.n_qubits, self.n_qubits))
        rotation_matrix_identical[0, 0] = 1
        rotation_matrix_identical[1, 1] = 1

        one_body = numpy.zeros((self.n_qubits, self.n_qubits))
        two_body = numpy.zeros((self.n_qubits, self.n_qubits,
                                self.n_qubits, self.n_qubits))
        one_body_spinful = numpy.zeros((2 * self.n_qubits, 2 * self.n_qubits))
        two_body_spinful = numpy.zeros((2 * self.n_qubits, 2 * self.n_qubits,
                                        2 * self.n_qubits, 2 * self.n_qubits))
        i = 0
        j = 0
        for p in range(self.n_qubits):
            for q in range(self.n_qubits):
                one_body[p, q] = i
                one_body_spinful[p, q] = i
                one_body_spinful[p + self.n_qubits, q + self.n_qubits] = i
                i = i + 1
                for r in range(self.n_qubits):
                    for s in range(self.n_qubits):
                        two_body[p, q, r, s] = j
                        two_body_spinful[p, q, r, s] = j
                        two_body_spinful[p + self.n_qubits,
                                         q + self.n_qubits,
                                         r + self.n_qubits,
                                         s + self.n_qubits] = j
                        j = j + 1
        polynomial_tensor = PolynomialTensor(
            {(): self.constant, (1, 0): one_body, (1, 1, 0, 0): two_body})
        want_polynomial_tensor = PolynomialTensor(
            {(): self.constant, (1, 0): one_body, (1, 1, 0, 0): two_body})
        polynomial_tensor_spinful = PolynomialTensor(
            {(): self.constant, (1, 0): one_body_spinful,
             (1, 1, 0, 0): two_body_spinful})
        want_polynomial_tensor_spinful = PolynomialTensor(
            {(): self.constant, (1, 0): one_body_spinful,
             (1, 1, 0, 0): two_body_spinful})

        polynomial_tensor.rotate_basis(rotation_matrix_identical)
        polynomial_tensor_spinful.rotate_basis(rotation_matrix_identical)
        self.assertEqual(polynomial_tensor, want_polynomial_tensor)
        self.assertEqual(polynomial_tensor_spinful,
                         want_polynomial_tensor_spinful)

    def test_rotate_basis_reverse(self):
        rotation_matrix_reverse = numpy.zeros((self.n_qubits, self.n_qubits))
        rotation_matrix_reverse[0, 1] = 1
        rotation_matrix_reverse[1, 0] = 1

        one_body = numpy.zeros((self.n_qubits, self.n_qubits))
        two_body = numpy.zeros((self.n_qubits, self.n_qubits,
                                self.n_qubits, self.n_qubits))
        one_body_reverse = numpy.zeros((self.n_qubits, self.n_qubits))
        two_body_reverse = numpy.zeros((self.n_qubits, self.n_qubits,
                                        self.n_qubits, self.n_qubits))
        i = 0
        j = 0
        i_reverse = pow(self.n_qubits, 2) - 1
        j_reverse = pow(self.n_qubits, 4) - 1
        for p in range(self.n_qubits):
            for q in range(self.n_qubits):
                one_body[p, q] = i
                i = i + 1
                one_body_reverse[p, q] = i_reverse
                i_reverse = i_reverse - 1
                for r in range(self.n_qubits):
                    for s in range(self.n_qubits):
                        two_body[p, q, r, s] = j
                        j = j + 1
                        two_body_reverse[p, q, r, s] = j_reverse
                        j_reverse = j_reverse - 1
        polynomial_tensor = PolynomialTensor(
            {(): self.constant, (1, 0): one_body, (1, 1, 0, 0): two_body})
        want_polynomial_tensor = PolynomialTensor(
            {(): self.constant, (1, 0): one_body_reverse,
             (1, 1, 0, 0): two_body_reverse})
        polynomial_tensor.rotate_basis(rotation_matrix_reverse)
        self.assertEqual(polynomial_tensor, want_polynomial_tensor)

    def test_rotate_basis_quadratic_hamiltonian_real(self):
        self.do_rotate_basis_quadratic_hamiltonian(True)

    def test_rotate_basis_quadratic_hamiltonian_complex(self):
        self.do_rotate_basis_quadratic_hamiltonian(False)

    def do_rotate_basis_quadratic_hamiltonian(self, real):
        """Test diagonalizing a quadratic Hamiltonian that conserves particle
        number."""
        n_qubits = 5

        # Initialize a particle-number-conserving quadratic Hamiltonian
        # and compute its orbital energies
        quad_ham = random_quadratic_hamiltonian(n_qubits, True, real=real)
        orbital_energies, constant = quad_ham.orbital_energies()

        # Rotate a basis where the Hamiltonian is diagonal
        hermitian_matrix = quad_ham.combined_hermitian_part
        energies, diagonalizing_unitary = numpy.linalg.eigh(hermitian_matrix)
        quad_ham.rotate_basis(diagonalizing_unitary)

        # Check that the rotated Hamiltonian is diagonal with the correct
        # orbital energies
        D = numpy.zeros((n_qubits, n_qubits), dtype=complex)
        D[numpy.diag_indices(n_qubits)] = orbital_energies
        self.assertTrue(numpy.allclose(quad_ham.combined_hermitian_part, D))

        # Check that the new Hamiltonian still conserves particle number
        self.assertTrue(quad_ham.conserves_particle_number)

        # Check that the orbital energies and constant are the same
        new_orbital_energies, new_constant = quad_ham.orbital_energies()
        self.assertTrue(numpy.allclose(orbital_energies, new_orbital_energies))
        self.assertAlmostEqual(constant, new_constant)

    def test_rotate_basis_max_order(self):
        for order in [15, 16]:
            tensor, want_tensor = self.do_rotate_basis_high_order(order)
            self.assertEqual(tensor, want_tensor)
        # I originally wanted to test 25 and 26, but it turns out that
        # numpy.einsum complains "too many subscripts in einsum" before 26.

        for order in [27, 28]:
            with self.assertRaises(ValueError):
                tensor, want_tensor = self.do_rotate_basis_high_order(order)

    def do_rotate_basis_high_order(self, order):
        key = (1,) * (order // 2) + (0,) * ((order + 1) // 2)
        shape = (1,) * order
        num = numpy.random.rand()
        rotation = numpy.exp(numpy.random.rand() * numpy.pi * 2j)

        polynomial_tensor = PolynomialTensor({key: numpy.zeros(shape) + num})

        # If order is odd, there are one more 0 than 1 in key
        if order % 2 == 1:
            num *= rotation
        want_polynomial_tensor = PolynomialTensor(
            {key: numpy.zeros(shape) + num})

        polynomial_tensor.rotate_basis(numpy.array([[rotation]]))

        return polynomial_tensor, want_polynomial_tensor



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""testing angular momentum generators. _fermion_spin_operators.py"""
import numpy
import unittest
from openfermion.ops import FermionOperator, normal_ordered
from openfermion.utils import commutator
from openfermion.utils._special_operators import (
        majorana_operator, number_operator,
        s_minus_operator, s_plus_operator, s_squared_operator,
        sx_operator, sy_operator, sz_operator,
        up_index, down_index)


class FermionSpinOperatorsTest(unittest.TestCase):

    def test_up_index(self):
        self.assertTrue(up_index(2) == 4)
        self.assertTrue(up_index(5) == 10)

    def test_up_down(self):
        self.assertTrue(down_index(2) == 5)
        self.assertTrue(down_index(5) == 11)

    def test_s_plus_operator(self):
        op = s_plus_operator(2)
        expected = (FermionOperator(((0, 1), (1, 0))) +
                    FermionOperator(((2, 1), (3, 0))))
        self.assertTrue(op == expected)

    def test_s_minus_operator(self):
        op = s_minus_operator(3)
        expected = (FermionOperator(((1, 1), (0, 0))) +
                    FermionOperator(((3, 1), (2, 0))) +
                    FermionOperator(((5, 1), (4, 0))))
        self.assertTrue(op == expected)

    def test_sx_operator(self):
        op = sx_operator(2)
        expected = (FermionOperator(((0, 1), (1, 0)), 0.5) +
                    FermionOperator(((1, 1), (0, 0)), 0.5) +
                    FermionOperator(((2, 1), (3, 0)), 0.5) +
                    FermionOperator(((3, 1), (2, 0)), 0.5))
        self.assertTrue(op == expected)

    def test_sy_operator(self):
        op = sy_operator(2)
        expected = (FermionOperator(((0, 1), (1, 0)), -0.5j) -
                    FermionOperator(((1, 1), (0, 0)), -0.5j) +
                    FermionOperator(((2, 1), (3, 0)), -0.5j) -
                    FermionOperator(((3, 1), (2, 0)), -0.5j))
        self.assertTrue(op == expected)

    def test_sz_operator(self):
        op = sz_operator(2)
        expected = (FermionOperator(((0, 1), (0, 0)), 0.5) -
                    FermionOperator(((1, 1), (1, 0)), 0.5) +
                    FermionOperator(((2, 1), (2, 0)), 0.5) -
                    FermionOperator(((3, 1), (3, 0)), 0.5))
        self.assertTrue(op == expected)

    def test_s_squared_operator(self):
        op = s_squared_operator(2)
        s_minus = (FermionOperator(((1, 1), (0, 0))) +
                   FermionOperator(((3, 1), (2, 0))))
        s_plus = (FermionOperator(((0, 1), (1, 0))) +
                  FermionOperator(((2, 1), (3, 0))))
        s_z = (FermionOperator(((0, 1), (0, 0)), 0.5) -
               FermionOperator(((1, 1), (1, 0)), 0.5) +
               FermionOperator(((2, 1), (2, 0)), 0.5) -
               FermionOperator(((3, 1), (3, 0)), 0.5))
        expected = s_minus * s_plus + s_z * s_z + s_z
        self.assertTrue(op == expected)

    def test_relations(self):
        n_spatial_orbitals = 2

        s_plus = s_plus_operator(n_spatial_orbitals)
        s_minus = s_minus_operator(n_spatial_orbitals)
        sx = sx_operator(n_spatial_orbitals)
        sy = sy_operator(n_spatial_orbitals)
        sz = sz_operator(n_spatial_orbitals)
        s_squared = s_squared_operator(n_spatial_orbitals)

        identity = FermionOperator(())

        self.assertEqual(normal_ordered(sx),
                         normal_ordered(.5 * (s_plus + s_minus)))
        self.assertEqual(normal_ordered(sy),
                         normal_ordered((.5 / 1.j) * (s_plus - s_minus)))
        self.assertEqual(normal_ordered(s_squared),
                         normal_ordered(sx ** 2 + sy ** 2 + sz ** 2))
        self.assertEqual(normal_ordered(s_squared),
                         normal_ordered(s_plus * s_minus +
                                        sz * (sz - identity)))
        self.assertEqual(normal_ordered(commutator(s_plus, s_minus)),
                         normal_ordered(2 * sz))
        self.assertEqual(normal_ordered(commutator(sx, sy)),
                         normal_ordered(1.j * sz))

    def test_invalid_input(self):

        with self.assertRaises(TypeError):
            s_minus_operator('a')

        with self.assertRaises(TypeError):
            s_plus_operator('a')

        with self.assertRaises(TypeError):
            sx_operator('a')

        with self.assertRaises(TypeError):
            sy_operator('a')

        with self.assertRaises(TypeError):
            sz_operator('a')

        with self.assertRaises(TypeError):
            s_squared_operator('a')


class NumberOperatorTest(unittest.TestCase):

    def test_number_operator_site(self):
        op = number_operator(3, 2, 1j)
        self.assertTrue(op == FermionOperator(((2, 1), (2, 0))) * 1j)

    def test_number_operator_nosite(self):
        op = number_operator(4)
        expected = (FermionOperator(((0, 1), (0, 0))) +
                    FermionOperator(((1, 1), (1, 0))) +
                    FermionOperator(((2, 1), (2, 0))) +
                    FermionOperator(((3, 1), (3, 0))))
        self.assertTrue(op == expected)


class MajoranaOperatorTest(unittest.TestCase):

    def test_init(self):
        # Test 'c' operator
        op1 = majorana_operator((2, 0))
        op2 = majorana_operator('c2')
        correct = FermionOperator('2^') + FermionOperator('2')
        self.assertTrue(op1 == op2)
        self.assertTrue(op1 == correct)

        # Test 'd' operator
        op1 = majorana_operator((3, 1))
        op2 = majorana_operator('d3')
        correct = FermionOperator('3^', 1.j) - FermionOperator('3', 1.j)
        self.assertTrue(op1 == op2)
        self.assertTrue(op1 == correct)

    def test_none_term(self):
        majorana_op = majorana_operator()
        self.assertTrue(majorana_operator() == FermionOperator())

    def test_bad_coefficient(self):
        with self.assertRaises(ValueError):
            majorana_op = majorana_operator((1, 1), 'a')

    def test_bad_term(self):
        with self.assertRaises(ValueError):
            majorana_op = majorana_operator((2, 2))
        with self.assertRaises(ValueError):
            majorana_op = majorana_operator('a')
        with self.assertRaises(ValueError):
            majorana_op = majorana_operator(2)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from ._binary_code_transform import (binary_code_transform,
                                     dissolve)
from ._bksf import bravyi_kitaev_fast
from ._binary_codes import (bravyi_kitaev_code,
                            checksum_code,
                            interleaved_code,
                            jordan_wigner_code,
                            parity_code,
                            weight_two_segment_code,
                            weight_one_binary_addressing_code,
                            weight_one_segment_code,
                            linearize_decoder)
from ._bravyi_kitaev import bravyi_kitaev
from ._bravyi_kitaev_tree import bravyi_kitaev_tree
from ._conversion import (get_fermion_operator,
                          get_interaction_rdm,
                          get_interaction_operator,
                          get_quadratic_hamiltonian,
                          get_molecular_data,
                          get_sparse_operator)
from ._jordan_wigner import jordan_wigner
from ._reverse_jordan_wigner import reverse_jordan_wigner
from ._verstraete_cirac import verstraete_cirac_2d_square



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""FermionOperator stores a sum of products of fermionic ladder operators."""
import numpy

from future.utils import iteritems
from openfermion.ops import SymbolicOperator


class FermionOperatorError(Exception):
    pass


def normal_ordered_term(term, coefficient):
    """Return a normal ordered FermionOperator corresponding to single term.

    Args:
        term: A tuple of tuples. The first element of each tuple is
            an integer indicating the mode on which a fermion ladder
            operator acts, starting from zero. The second element of each
            tuple is an integer, either 1 or 0, indicating whether creation
            or annihilation acts on that mode.
        coefficient: The coefficient of the term.

    Returns:
        ordered_term (FermionOperator): The normal ordered form of the input.
            Note that this might have more terms.

    In our convention, normal ordering implies terms are ordered
    from highest tensor factor (on left) to lowest (on right).
    Also, ladder operators come first.

    Warning:
        Even assuming that each creation or annihilation operator appears
        at most a constant number of times in the original term, the
        runtime of this method is exponential in the number of qubits.
    """
    # Iterate from left to right across operators and reorder to normal
    # form. Swap terms operators into correct position by moving from
    # left to right across ladder operators.
    term = list(term)
    ordered_term = FermionOperator()
    for i in range(1, len(term)):
        for j in range(i, 0, -1):
            right_operator = term[j]
            left_operator = term[j - 1]

            # Swap operators if raising on right and lowering on left.
            if right_operator[1] and not left_operator[1]:
                term[j - 1] = right_operator
                term[j] = left_operator
                coefficient *= -1

                # Replace a a^\dagger with 1 - a^\dagger a
                # if indices are the same.
                if right_operator[0] == left_operator[0]:
                    new_term = term[:(j - 1)] + term[(j + 1)::]

                    # Recursively add the processed new term.
                    ordered_term += normal_ordered_term(
                        tuple(new_term), -coefficient)

            # Handle case when operator type is the same.
            elif right_operator[1] == left_operator[1]:

                # If same two operators are repeated, evaluate to zero.
                if right_operator[0] == left_operator[0]:
                    return ordered_term

                # Swap if same ladder type but lower index on left.
                elif right_operator[0] > left_operator[0]:
                    term[j - 1] = right_operator
                    term[j] = left_operator
                    coefficient *= -1

    # Add processed term and return.
    ordered_term += FermionOperator(tuple(term), coefficient)
    return ordered_term


def normal_ordered(fermion_operator):
    """Compute and return the normal ordered form of a FermionOperator.

    In our convention, normal ordering implies terms are ordered
    from highest tensor factor (on left) to lowest (on right).
    Also, ladder operators come first.

    Warning:
        Even assuming that each creation or annihilation operator appears
        at most a constant number of times in the original term, the
        runtime of this method is exponential in the number of qubits.
    """
    ordered_operator = FermionOperator()
    for term, coefficient in fermion_operator.terms.items():
        ordered_operator += normal_ordered_term(term, coefficient)
    return ordered_operator


class FermionOperator(SymbolicOperator):
    """FermionOperator stores a sum of products of fermionic ladder operators.

    In OpenFermion, we describe fermionic ladder operators using the shorthand:
    'q^' = a^\dagger_q
    'q' = a_q
    where {'p^', 'q'} = delta_pq

    One can multiply together these fermionic ladder operators to obtain a
    fermionic term. For instance, '2^ 1' is a fermion term which
    creates at orbital 2 and destroys at orbital 1. The FermionOperator class
    also stores a coefficient for the term, e.g. '3.17 * 2^ 1'.

    The FermionOperator class is designed (in general) to store sums of these
    terms. For instance, an instance of FermionOperator might represent
    3.17 2^ 1 - 66.2 * 8^ 7 6^ 2
    The Fermion Operator class overloads operations for manipulation of
    these objects by the user.

    FermionOperator is a subclass of SymbolicOperator. Importantly, it has
    attributes set as follows::

        actions = (1, 0)
        action_strings = ('^', '')
        action_before_index = False
        different_indices_commute = False

    See the documentation of SymbolicOperator for more details.

    Example:
        .. code-block:: python

            ham = (FermionOperator('0^ 3', .5)
                   + .5 * FermionOperator('3^ 0'))
            # Equivalently
            ham2 = FermionOperator('0^ 3', 0.5)
            ham2 += FermionOperator('3^ 0', 0.5)

    Note:
        Adding FermionOperators is faster using += (as this
        is done by in-place addition). Specifying the coefficient
        during initialization is faster than multiplying a FermionOperator
        with a scalar.
    """
    actions = (1, 0)
    action_strings = ('^', '')
    action_before_index = False
    different_indices_commute = False

    def is_normal_ordered(self):
        """Return whether or not term is in normal order.

        In our convention, normal ordering implies terms are ordered
        from highest tensor factor (on left) to lowest (on right). Also,
        ladder operators come first.
        """
        for term in self.terms:
            for i in range(1, len(term)):
                for j in range(i, 0, -1):
                    right_operator = term[j]
                    left_operator = term[j - 1]
                    if right_operator[1] and not left_operator[1]:
                        return False
                    elif (right_operator[1] == left_operator[1] and
                          right_operator[0] >= left_operator[0]):
                        return False
        return True

    def is_molecular_term(self):
        """Query whether term has correct form to be from a molecular.

        Require that term is particle-number conserving (same number of
        raising and lowering operators). Require that term has 0, 2 or 4
        ladder operators. Require that term conserves spin (parity of
        raising operators equals parity of lowering operators).
        """
        for term in self.terms:
            if len(term) not in (0, 2, 4):
                return False

            # Make sure term conserves particle number and spin.
            spin = 0
            particles = 0
            for operator in term:
                particles += (-1) ** operator[1]  # add 1 if create, else -1
                spin += (-1) ** (operator[0] + operator[1])
            if not (particles == spin == 0):
                return False
        return True



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Tests for _verstraete_cirac.py."""
from __future__ import absolute_import

import unittest

import numpy

from openfermion.hamiltonians import fermi_hubbard
from openfermion.transforms import (get_sparse_operator,
                                    verstraete_cirac_2d_square)
from openfermion.transforms._verstraete_cirac import (
    coordinates_to_snake_index, snake_index_to_coordinates,
    stabilizer_local_2d_square)
from openfermion.utils import get_ground_state


class VerstraeteCirac2dSquareGroundStateTest(unittest.TestCase):
    """Test that the transform preserves desired ground state properties."""

    def setUp(self):
        self.x_dimension = 2
        self.y_dimension = 3

        # Create a Hamiltonian with nearest-neighbor hopping terms
        self.ferm_op = fermi_hubbard(self.x_dimension, self.y_dimension,
                                     1., 0., 0., 0., False, True)

        # Get the ground energy and ground state
        self.ferm_op_sparse = get_sparse_operator(self.ferm_op)
        self.ferm_op_ground_energy, self.ferm_op_ground_state = (
            get_ground_state(self.ferm_op_sparse))

        # Transform the FermionOperator to a QubitOperator
        self.transformed_op = verstraete_cirac_2d_square(
            self.ferm_op, self.x_dimension, self.y_dimension,
            add_auxiliary_hamiltonian=True, snake=False)

        # Get the ground energy and state of the transformed operator
        self.transformed_sparse = get_sparse_operator(self.transformed_op)
        self.transformed_ground_energy, self.transformed_ground_state = (
            get_ground_state(self.transformed_sparse))

    def test_ground_energy(self):
        """Test that the transformation preserves the ground energy."""
        self.assertAlmostEqual(self.transformed_ground_energy,
                               self.ferm_op_ground_energy)


class VerstraeteCirac2dSquareOperatorLocalityTest(unittest.TestCase):
    """Test that the transform results in local qubit operators."""

    def setUp(self):
        self.x_dimension = 6
        self.y_dimension = 6

        # Create a Hubbard Hamiltonian
        self.ferm_op = fermi_hubbard(self.x_dimension, self.y_dimension,
                                     1.0, 4.0, 0.0, 0.0, False, True)

        # Transform the FermionOperator to a QubitOperator without including
        # the auxiliary Hamiltonian
        self.transformed_op_no_aux = verstraete_cirac_2d_square(
            self.ferm_op, self.x_dimension, self.y_dimension,
            add_auxiliary_hamiltonian=False, snake=False)
        self.transformed_op_no_aux.compress()

        # Transform the FermionOperator to a QubitOperator, including
        # the auxiliary Hamiltonian
        self.transformed_op_aux = verstraete_cirac_2d_square(
            self.ferm_op, self.x_dimension, self.y_dimension,
            add_auxiliary_hamiltonian=True, snake=False)
        self.transformed_op_aux.compress()

    def test_operator_locality_no_aux(self):
        """Test that the operators without the auxiliary Hamiltonian
        are at most 4-local."""
        for term in self.transformed_op_no_aux.terms:
            self.assertTrue(len(term) <= 4)

    def test_operator_locality_aux(self):
        """Test that the operators with the auxiliary Hamiltonian
        are at most 6-local."""
        for term in self.transformed_op_aux.terms:
            self.assertTrue(len(term) <= 6)


class ExceptionTest(unittest.TestCase):
    """Test that exceptions are raised correctly."""

    def test_verstraete_cirac_2d_square(self):
        ferm_op = fermi_hubbard(3, 2, 1., 0., spinless=True)
        with self.assertRaises(NotImplementedError):
            operator = verstraete_cirac_2d_square(ferm_op, 3, 2)

    def test_stabilizer_local_2d_square(self):
        with self.assertRaises(ValueError):
            index = stabilizer_local_2d_square(0, 2, 4, 4)

    def test_coordinates_to_snake_index(self):
        with self.assertRaises(ValueError):
            index = coordinates_to_snake_index(4, 4, 4, 5)
        with self.assertRaises(ValueError):
            index = coordinates_to_snake_index(4, 4, 5, 4)

    def test_snake_index_to_coordinates(self):
        with self.assertRaises(ValueError):
            row, col = snake_index_to_coordinates(20, 4, 5)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Tests for _lcu_util.py."""

from __future__ import absolute_import, division

import random
import unittest

from openfermion.utils._lcu_util import (
    _preprocess_for_efficient_roulette_selection,
    _discretize_probability_distribution,
    preprocess_lcu_coefficients_for_reversible_sampling,
)


class DiscretizeDistributionTest(unittest.TestCase):

    def assertGetDiscretizedDistribution(self, probabilities, epsilon):
        total_probability = sum(probabilities)
        numers, denom = _discretize_probability_distribution(probabilities,
                                                             epsilon)
        self.assertEqual(sum(numers), denom)
        self.assertEqual(len(numers), len(probabilities))
        for i in range(len(numers)):
            self.assertAlmostEqual(numers[i] / denom,
                                   probabilities[i] / total_probability,
                                   delta=epsilon)
        return numers, denom

    def test_fuzz(self):
        for _ in range(100):
            n = random.randint(1, 50)
            weights = [random.random() for _ in range(n)]
            self.assertGetDiscretizedDistribution(
                weights, 2**-random.randint(1, 20))

    def test_known_discretizations(self):
        self.assertEqual(
            self.assertGetDiscretizedDistribution([1], 0.25),
            ([4], 4))

        self.assertEqual(
            self.assertGetDiscretizedDistribution([1], 0.125),
            ([8], 8))

        self.assertEqual(
            self.assertGetDiscretizedDistribution([0.1, 0.1, 0.1], 0.25),
            ([2, 2, 2], 6))

        self.assertEqual(
            self.assertGetDiscretizedDistribution([0.09, 0.11, 0.1], 0.25),
            ([2, 2, 2], 6))

        self.assertEqual(
            self.assertGetDiscretizedDistribution([0.09, 0.11, 0.1], 0.1),
            ([4, 4, 4], 12))

        self.assertEqual(
            self.assertGetDiscretizedDistribution([0.09, 0.11, 0.1], 0.05),
            ([6, 8, 7], 21))

        self.assertEqual(
            self.assertGetDiscretizedDistribution([0.09, 0.11, 0.1], 0.01),
            ([31, 37, 34], 102))

        self.assertEqual(
            self.assertGetDiscretizedDistribution([0.09, 0.11, 0.1], 0.00335),
            ([90, 110, 100], 300))


class PreprocessForEfficientRouletteSelectionTest(unittest.TestCase):

    def assertPreprocess(self, weights):
        alternates, keep_chances = (
            _preprocess_for_efficient_roulette_selection(weights))

        self.assertEqual(len(alternates), len(keep_chances))

        target_weight = sum(weights) // len(alternates)
        distribution = list(keep_chances)
        for i in range(len(alternates)):
            distribution[alternates[i]] += target_weight - keep_chances[i]
        self.assertEqual(weights, distribution)

        return alternates, keep_chances

    def test_fuzz(self):
        for _ in range(100):
            n = random.randint(1, 50)
            weights = [random.randint(0, 100) for _ in range(n)]
            weights[-1] += n - sum(weights) % n  # Ensure multiple of length.
            self.assertPreprocess(weights)

    def test_validation(self):
        with self.assertRaises(ValueError):
            _ = self.assertPreprocess(weights=[])
        with self.assertRaises(ValueError):
            _ = self.assertPreprocess(weights=[1, 2])
        with self.assertRaises(ValueError):
            _ = self.assertPreprocess(weights=[3, 3, 2])

    def test_already_uniform(self):
        self.assertEqual(
            self.assertPreprocess(weights=[1]),
            ([0], [0]))
        self.assertEqual(
            self.assertPreprocess(weights=[1, 1]),
            ([0, 1], [0, 0]))
        self.assertEqual(
            self.assertPreprocess(weights=[1, 1, 1]),
            ([0, 1, 2], [0, 0, 0]))
        self.assertEqual(
            self.assertPreprocess(weights=[2, 2, 2]),
            ([0, 1, 2], [0, 0, 0]))

    def test_donation(self):
        # v2 donates 1 to v0.
        self.assertEqual(
            self.assertPreprocess(weights=[1, 2, 3]),
            ([2, 1, 2], [1, 0, 0]))
        # v0 donates 1 to v1.
        self.assertEqual(
            self.assertPreprocess(weights=[3, 1, 2]),
            ([0, 0, 2], [0, 1, 0]))
        # v0 donates 1 to v1, then 2 to v2.
        self.assertEqual(
            self.assertPreprocess(weights=[5, 1, 0]),
            ([0, 0, 0], [0, 1, 0]))

    def test_over_donation(self):
        # v0 donates 2 to v1, leaving v0 needy, then v2 donates 1 to v0.
        self.assertEqual(
            self.assertPreprocess(weights=[3, 0, 3]),
            ([2, 0, 2], [1, 0, 0]))


class PreprocessLCUCoefficientsForReversibleSamplingTest(unittest.TestCase):
    def assertPreprocess(self, lcu_coefs, epsilon):
        alternates, keep_numers, keep_denom = (
            preprocess_lcu_coefficients_for_reversible_sampling(
                lcu_coefs, epsilon))

        n = len(lcu_coefs)
        self.assertEqual(len(alternates), n)
        self.assertEqual(len(keep_numers), n)

        out_distribution = [numer/keep_denom for numer in keep_numers]
        for i in range(n):
            out_distribution[alternates[i]] += 1/n - keep_numers[i]/keep_denom

        total = sum(lcu_coefs)
        for i in range(n):
            self.assertAlmostEqual(out_distribution[i],
                                   lcu_coefs[i] / total,
                                   delta=epsilon)

        return alternates, keep_numers, keep_denom

    def test_fuzz(self):
        for _ in range(100):
            n = random.randint(1, 50)
            weights = [random.randint(0, 100) for _ in range(n)]
            weights[-1] += n - sum(weights) % n  # Ensure multiple of length.
            self.assertPreprocess(weights, 2**-random.randint(1, 20))

    def test_known(self):
        self.assertEqual(
            self.assertPreprocess([1, 2], epsilon=0.01),
            ([1, 1], [33, 0], 100))



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Module to compute Trotter errors in the plane-wave dual basis."""
from __future__ import absolute_import
from future.utils import iteritems, itervalues

import numpy

from openfermion.config import *
from openfermion.hamiltonians import jellium_model, wigner_seitz_length_scale
from openfermion.ops import FermionOperator, normal_ordered
from openfermion.utils import count_qubits, Grid
from openfermion.utils._commutators import (
    double_commutator,
    trivially_double_commutes_dual_basis,
    trivially_double_commutes_dual_basis_using_term_info)


def low_depth_second_order_trotter_error_operator(
        terms, indices=None, is_hopping_operator=None, jellium_only=False,
        verbose=False):
    """Determine the difference between the exact generator of unitary
    evolution and the approximate generator given by the second-order
    Trotter-Suzuki expansion.

    Args:
        terms: a list of FermionOperators in the Hamiltonian in the
               order in which they will be simulated.
        indices: a set of indices the terms act on in the same order as terms.
        is_hopping_operator: a list of whether each term is a hopping operator.
        jellium_only: Whether the terms are from the jellium Hamiltonian only,
                      rather than the full dual basis Hamiltonian (i.e. whether
                      c_i = c for all number operators i^ i, or whether they
                      depend on i as is possible in the general case).
        verbose: Whether to print percentage progress.

    Returns:
        The difference between the true and effective generators of time
            evolution for a single Trotter step.

    Notes: follows Equation 9 of Poulin et al.'s work in "The Trotter Step
        Size Required for Accurate Quantum Simulation of Quantum Chemistry",
        applied to the "stagger"-based Trotter step for detailed in
        Kivlichan et al., "Quantum Simulation of Electronic Structure with
        Linear Depth and Connectivity", arxiv:1711.04789.
    """
    more_info = bool(indices)
    n_terms = len(terms)

    if verbose:
        import time
        start = time.time()

    error_operator = FermionOperator.zero()
    for beta in range(n_terms):
        if verbose and beta % (n_terms // 30) == 0:
            print('%4.3f percent done in' % (
                (float(beta) / n_terms) ** 3 * 100), time.time() - start)

        for alpha in range(beta + 1):
            for alpha_prime in range(beta):
                # If we have pre-computed info on indices, use it to determine
                # trivial double commutation.
                if more_info:
                    if (not
                        trivially_double_commutes_dual_basis_using_term_info(
                            indices[alpha], indices[beta],
                            indices[alpha_prime], is_hopping_operator[alpha],
                            is_hopping_operator[beta],
                            is_hopping_operator[alpha_prime], jellium_only)):
                        # Determine the result of the double commutator.
                        double_com = double_commutator(
                            terms[alpha], terms[beta], terms[alpha_prime],
                            indices[beta], indices[alpha_prime],
                            is_hopping_operator[beta],
                            is_hopping_operator[alpha_prime])
                        if alpha == beta:
                            double_com /= 2.0

                        error_operator += double_com

                # If we don't have more info, check for trivial double
                # commutation using the terms directly.
                elif not trivially_double_commutes_dual_basis(
                        terms[alpha], terms[beta], terms[alpha_prime]):
                    double_com = double_commutator(
                        terms[alpha], terms[beta], terms[alpha_prime])

                    if alpha == beta:
                        double_com /= 2.0

                    error_operator += double_com

    error_operator /= 12.0
    return error_operator


def low_depth_second_order_trotter_error_bound(
        terms, indices=None, is_hopping_operator=None,
        jellium_only=False, verbose=False):
    """Numerically upper bound the error in the ground state energy
    for the second-order Trotter-Suzuki expansion.

    Args:
        terms: a list of single-term FermionOperators in the Hamiltonian
            to be simulated.
        indices: a set of indices the terms act on in the same order as terms.
        is_hopping_operator: a list of whether each term is a hopping operator.
        jellium_only: Whether the terms are from the jellium Hamiltonian only,
                      rather than the full dual basis Hamiltonian (i.e. whether
                      c_i = c for all number operators i^ i, or whether they
                      depend on i as is possible in the general case).
        verbose: Whether to print percentage progress.

    Returns:
        A float upper bound on norm of error in the ground state energy.

    Notes:
        Follows Equation 9 of Poulin et al.'s work in "The Trotter Step
        Size Required for Accurate Quantum Simulation of Quantum
        Chemistry" to calculate the error operator, for the "stagger"-based
        Trotter step for detailed in Kivlichan et al., "Quantum Simulation
        of Electronic Structure with Linear Depth and Connectivity",
        arxiv:1711.04789.
    """
    # Return the 1-norm of the error operator (upper bound on error).
    return numpy.sum(numpy.absolute(list(
        low_depth_second_order_trotter_error_operator(
            terms, indices, is_hopping_operator,
            jellium_only, verbose).terms.values())))


def simulation_ordered_grouped_low_depth_terms_with_info(
        hamiltonian, input_ordering=None):
    """Give terms from the dual basis Hamiltonian in simulated order.

    Uses the simulation ordering, grouping terms into hopping
    (i^ j + j^ i) and number (i^j^ i j + c_i i^ i + c_j j^ j) operators.
    Pre-computes term information (indices each operator acts on, as
    well as whether each operator is a hopping operator.

    Args:
        hamiltonian (FermionOperator): The Hamiltonian.
        input_ordering (list): The initial Jordan-Wigner canonical order.

    Returns:
        A 3-tuple of terms from the Hamiltonian in order of simulation,
        the indices they act on, and whether they are hopping operators
        (both also in the same order).

    Notes:
        Follows the "stagger"-based simulation order discussed in Kivlichan
        et al., "Quantum Simulation of Electronic Structure with Linear
        Depth and Connectivity", arxiv:1711.04789; as such, the only
        permitted types of terms are hopping (i^ j + j^ i) and potential
        terms which are products of at most two number operators.
    """
    n_qubits = count_qubits(hamiltonian)
    hamiltonian = normal_ordered(hamiltonian)

    ordered_terms = []
    ordered_indices = []
    ordered_is_hopping_operator = []

    # If no input mode ordering is specified, default to range(n_qubits).
    try:
        input_ordering = list(input_ordering)
    except TypeError:
        input_ordering = list(range(n_qubits))

    # Half a second-order Trotter step reverses the input ordering: this tells
    # us how much we need to include in the ordered list of terms.
    final_ordering = list(reversed(input_ordering))

    # Follow odd-even transposition sort. In alternating steps, swap each even
    # qubits with the odd qubit to its right, and in the next step swap each
    # the odd qubits with the even qubit to its right. Do this until the input
    # ordering has been reversed.
    parity = 0
    while input_ordering != final_ordering:
        results = stagger_with_info(
            hamiltonian, input_ordering, parity)
        terms_in_step, indices_in_step, is_hopping_operator_in_step = results

        ordered_terms.extend(terms_in_step)
        ordered_indices.extend(indices_in_step)
        ordered_is_hopping_operator.extend(is_hopping_operator_in_step)

        # Alternate even and odd steps of the reversal procedure.
        parity = 1 - parity

    return (ordered_terms, ordered_indices, ordered_is_hopping_operator)


def stagger_with_info(hamiltonian, input_ordering, parity):
    """Give terms simulated in a single stagger of a Trotter step.

    Groups terms into hopping (i^ j + j^ i) and number
    (i^j^ i j + c_i i^ i + c_j j^ j) operators.
    Pre-computes term information (indices each operator acts on, as
    well as whether each operator is a hopping operator).

    Args:
        hamiltonian (FermionOperator): The Hamiltonian.
        input_ordering (list): The initial Jordan-Wigner canonical order.
        parity (boolean): Whether to determine the terms from the next even
            (False = 0) or odd (True = 1) stagger.

    Returns:
        A 3-tuple of terms from the Hamiltonian that are simulated in the
        stagger, the indices they act on, and whether they are hopping
        operators (all in the same order).

    Notes:
        The "staggers" used here are the left (parity=False) and right
        (parity=True) staggers detailed in Kivlichan et al., "Quantum
        Simulation of Electronic Structure with Linear Depth and
        Connectivity", arxiv:1711.04789. As such, the Hamiltonian must be
        in the form discussed in that paper. This constrains it to have
        only hopping terms (i^ j + j^ i) and potential terms which are
        products of at most two number operators (n_i or n_i n_j).
    """
    terms_in_step = []
    indices_in_step = []
    is_hopping_operator_in_step = []

    zero = FermionOperator.zero()
    n_qubits = count_qubits(hamiltonian)

    # A single round of odd-even transposition sort.
    for i in range(parity, n_qubits - 1, 2):
        # Always keep the max on the left to avoid having to normal order.
        left = max(input_ordering[i], input_ordering[i + 1])
        right = min(input_ordering[i], input_ordering[i + 1])

        # Calculate the hopping operators in the Hamiltonian.
        left_hopping_operator = FermionOperator(
            ((left, 1), (right, 0)), hamiltonian.terms.get(
                ((left, 1), (right, 0)), 0.0))
        right_hopping_operator = FermionOperator(
            ((right, 1), (left, 0)), hamiltonian.terms.get(
                ((right, 1), (left, 0)), 0.0))

        # Calculate the two-number operator l^ r^ l r in the Hamiltonian.
        two_number_operator = FermionOperator(
            ((left, 1), (right, 1), (left, 0), (right, 0)),
            hamiltonian.terms.get(
                ((left, 1), (right, 1), (left, 0), (right, 0)), 0.0))

        # Calculate the left number operator, left^ left.
        left_number_operator = FermionOperator(
            ((left, 1), (left, 0)), hamiltonian.terms.get(
                ((left, 1), (left, 0)), 0.0))

        # Calculate the right number operator, right^ right.
        right_number_operator = FermionOperator(
            ((right, 1), (right, 0)), hamiltonian.terms.get(
                ((right, 1), (right, 0)), 0.0))

        # Divide single-number terms by n_qubits-1 to avoid over-counting.
        # Each qubit is swapped n_qubits-1 times total.
        left_number_operator /= (n_qubits - 1)
        right_number_operator /= (n_qubits - 1)

        # If the overall hopping operator isn't close to zero, append it.
        # Include the indices it acts on and that it's a hopping operator.
        if not (left_hopping_operator +
                right_hopping_operator) == zero:
            terms_in_step.append(left_hopping_operator +
                                 right_hopping_operator)
            indices_in_step.append(set((left, right)))
            is_hopping_operator_in_step.append(True)

        # If the overall number operator isn't close to zero, append it.
        # Include the indices it acts on and that it's a number operator.
        if not (two_number_operator + left_number_operator +
                right_number_operator) == zero:
            terms_in_step.append(two_number_operator +
                                 left_number_operator +
                                 right_number_operator)
            indices_in_step.append(set((left, right)))
            is_hopping_operator_in_step.append(False)

        # Modify the current Jordan-Wigner canonical ordering in-place.
        input_ordering[i], input_ordering[i + 1] = (input_ordering[i + 1],
                                                    input_ordering[i])

    return terms_in_step, indices_in_step, is_hopping_operator_in_step


def ordered_low_depth_terms_no_info(hamiltonian):
    """Give terms from Hamiltonian in dictionary output order.

    Args:
        hamiltonian (FermionOperator): The Hamiltonian.

    Returns:
        A list of terms from the Hamiltonian in simulated order.

    Notes:
        Assumes the Hamiltonian is in the form discussed in Kivlichan
        et al., "Quantum Simulation of Electronic Structure with Linear
        Depth and Connectivity", arxiv:1711.04789. This constrains the
        Hamiltonian to have only hopping terms (i^ j + j^ i) and potential
        terms which are products of at most two number operators (n_i or
        n_i n_j).
    """
    n_qubits = count_qubits(hamiltonian)
    hamiltonian = normal_ordered(hamiltonian)
    terms = []

    for operators, coefficient in iteritems(hamiltonian.terms):
        terms += [FermionOperator(operators, coefficient)]

    return terms


def dual_basis_jellium_hamiltonian(grid_length, dimension=3,
                                   wigner_seitz_radius=10., n_particles=None,
                                   spinless=True):
    """Return the jellium Hamiltonian with the given parameters.

    Args:
        grid_length (int): The number of spatial orbitals per dimension.
        dimension (int): The dimension of the system.
        wigner_seitz_radius (float): The radius per particle in Bohr.
        n_particles (int): The number of particles in the system.
                           Defaults to half filling if not specified.
    """
    n_qubits = grid_length ** dimension
    if not spinless:
        n_qubits *= 2

    if n_particles is None:
        # Default to half filling fraction.
        n_particles = n_qubits // 2

    if not (0 <= n_particles <= n_qubits):
        raise ValueError('n_particles must be between 0 and the number of'
                         ' spin-orbitals.')

    # Compute appropriate length scale.
    length_scale = wigner_seitz_length_scale(
        wigner_seitz_radius, n_particles, dimension)

    grid = Grid(dimension, grid_length, length_scale)
    hamiltonian = jellium_model(grid, spinless=spinless, plane_wave=False)
    hamiltonian = normal_ordered(hamiltonian)
    hamiltonian.compress()
    return hamiltonian



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Module to compute Trotter errors in the plane-wave dual basis."""
from __future__ import absolute_import

import numpy

from openfermion.config import *
from openfermion.ops import FermionOperator, normal_ordered
from openfermion.utils import count_qubits
from openfermion.utils._low_depth_trotter_error import stagger_with_info


def simulation_ordered_grouped_hubbard_terms_with_info(
        hubbard_hamiltonian):
    """Give terms from the Fermi-Hubbard Hamiltonian in simulated order.

    Uses the simulation ordering, grouping terms into hopping
    (i^ j + j^ i) and on-site potential (i^j^ i j) operators.
    Pre-computes term information (indices each operator acts on, as
    well as whether each operator is a hopping operator).

    Args:
        hubbard_hamiltonian (FermionOperator): The Hamiltonian.
        original_ordering (list): The initial Jordan-Wigner canonical order.

    Returns:
        A 3-tuple of terms from the Hubbard Hamiltonian in order of
        simulation, the indices they act on, and whether they are hopping
        operators (both also in the same order).

    Notes:
        Assumes that the Hubbard model has spin and is on a 2D square
        aperiodic lattice. Uses the "stagger"-based Trotter step for the
        Hubbard model detailed in Kivlichan et al., "Quantum Simulation
        of Electronic Structure with Linear Depth and Connectivity",
        arxiv:1711.04789.
    """
    hamiltonian = normal_ordered(hubbard_hamiltonian)
    n_qubits = count_qubits(hamiltonian)
    side_length = int(numpy.sqrt(n_qubits / 2.0))

    ordered_terms = []
    ordered_indices = []
    ordered_is_hopping_operator = []

    original_ordering = list(range(n_qubits))
    for i in range(0, n_qubits - side_length, 2 * side_length):
        for j in range(2 * bool(i % (4 * side_length)), 2 * side_length, 4):
            original_ordering[i+j], original_ordering[i+j+1] = (
                original_ordering[i+j+1], original_ordering[i+j])

    input_ordering = list(original_ordering)

    # Follow odd-even transposition sort. In alternating steps, swap each even
    # qubits with the odd qubit to its right, and in the next step swap each
    # the odd qubits with the even qubit to its right. Do this until the input
    # ordering has been reversed.
    parities = [False, True] * int(side_length / 2 + 1)  # int floors here
    for parity in parities:
        results = stagger_with_info(
            hubbard_hamiltonian, input_ordering, parity)
        terms_in_step, indices_in_step, is_hopping_operator_in_step = results

        ordered_terms.extend(terms_in_step)
        ordered_indices.extend(indices_in_step)
        ordered_is_hopping_operator.extend(is_hopping_operator_in_step)

    input_ordering = list(original_ordering)

    parities = [True] + [False, True] * int(side_length / 2)  # int floors here
    for parity in parities:
        results = stagger_with_info(
            hubbard_hamiltonian, input_ordering, parity)
        terms_in_step, indices_in_step, is_hopping_operator_in_step = results

        ordered_terms.extend(terms_in_step)
        ordered_indices.extend(indices_in_step)
        ordered_is_hopping_operator.extend(is_hopping_operator_in_step)

    return (ordered_terms, ordered_indices, ordered_is_hopping_operator)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Module to manipulate basic models of quantum channels"""

from functools import reduce
from itertools import chain
from numpy import array, conj, dot, eye, kron, log2, sqrt


def _verify_channel_inputs(density_matrix, probability, target_qubit):
    """Verifies input parameters for channels

    Args:
        density_matrix (numpy.ndarray): Density matrix of the system
        probability (float): Probability error is applied p \in [0, 1]
        target_qubit (int): target for the channel error.

    Returns:
        new_density_matrix(numpy.ndarray): Density matrix with the channel
            applied.
    """
    n_qubits = int(log2(density_matrix.shape[0]))

    if (len(density_matrix.shape) != 2 or
            density_matrix.shape[0] != density_matrix.shape[1]):
        raise ValueError("Error in input of density matrix to channel.")
    if (probability < 0) or (probability > 1):
        raise ValueError("Channel probability must be between 0 and 1.")
    if (target_qubit < 0) or (target_qubit >= n_qubits):
        raise ValueError("Target qubits must be within number of qubits.")


def _lift_operator(operator, n_qubits, target_qubit):
    """Lift a single qubit operator into the n_qubit space by kron product

    Args:
        operator (ndarray): Single qubit operator to lift into full space
        n_qubits (int): Number of total qubits in the space
        target_qubit (int): Qubit to act on

    Return:
        new_operator(Sparse Operator): Operator representing the embedding in
            the full space.
    """
    new_operator = (
        reduce(kron,
               chain(
                   (eye(2) for i in range(0, target_qubit)),
                   [operator],
                   (eye(2) for i in range(target_qubit + 1, n_qubits)))))
    return new_operator


def amplitude_damping_channel(density_matrix, probability, target_qubit,
                              transpose=False):
    """Apply an amplitude damping channel

    Applies an amplitude damping channel with a given probability to the target
    qubit in the density_matrix.

    Args:
        density_matrix (numpy.ndarray): Density matrix of the system
        probability (float): Probability error is applied p \in [0, 1]
        target_qubit (int): target for the channel error.
        transpose (bool): Conjugate transpose channel operators, useful for
            acting on Hamiltonians in variational channel state models

    Returns:
        new_density_matrix(numpy.ndarray): Density matrix with the channel
            applied.
    """
    _verify_channel_inputs(density_matrix, probability, target_qubit)
    n_qubits = int(log2(density_matrix.shape[0]))

    E0 = _lift_operator(array([[1.0, 0.0],
                               [0.0, sqrt(1.0 - probability)]], dtype=complex),
                        n_qubits, target_qubit)

    E1 = _lift_operator(array([[0.0, sqrt(probability)],
                               [0.0, 0.0]], dtype=complex),
                        n_qubits, target_qubit)

    if transpose:
        E0 = E0.T
        E1 = E1.T

    new_density_matrix = (dot(E0, dot(density_matrix, E0.T)) +
                          dot(E1, dot(density_matrix, E1.T)))

    return new_density_matrix


def dephasing_channel(density_matrix, probability, target_qubit,
                      transpose=False):
    """Apply a dephasing channel

    Applies an amplitude damping channel with a given probability to the target
    qubit in the density_matrix.

    Args:
        density_matrix (numpy.ndarray): Density matrix of the system
        probability (float): Probability error is applied p \in [0, 1]
        target_qubit (int): target for the channel error.
        transpose (bool): Conjugate transpose channel operators, useful for
            acting on Hamiltonians in variational channel state models

    Returns:
        new_density_matrix (numpy.ndarray): Density matrix with the channel
            applied.
    """
    _verify_channel_inputs(density_matrix, probability, target_qubit)
    n_qubits = int(log2(density_matrix.shape[0]))

    E0 = _lift_operator(sqrt(1.0 - probability/2.) * eye(2),
                        n_qubits, target_qubit)
    E1 = _lift_operator(sqrt(probability/2.) *
                        array([[1.0, 0.0], [1.0, -1.0]]),
                        n_qubits, target_qubit)

    if transpose:
        E0 = E0.T
        E1 = E1.T

    new_density_matrix = (dot(E0, dot(density_matrix, E0.T)) +
                          dot(E1, dot(density_matrix, E1.T)))

    return new_density_matrix


def depolarizing_channel(density_matrix, probability, target_qubit,
                         transpose=False):
    """Apply a depolarizing channel

    Applies an amplitude damping channel with a given probability to the target
    qubit in the density_matrix.

    Args:
        density_matrix (numpy.ndarray): Density matrix of the system
        probability (float): Probability error is applied p \in [0, 1]
        target_qubit (int/str): target for the channel error, if given special
            value "all", then a total depolarizing channel is applied.
        transpose (bool): Dummy parameter to match signature of other
            channels but depolarizing channel is symmetric under
            conjugate transpose.

    Returns:
        new_density_matrix (numpy.ndarray): Density matrix with the channel
            applied.
    """
    n_qubits = int(log2(density_matrix.shape[0]))

    # Toggle depolarizing channel on all qubits
    if isinstance(target_qubit, str) and target_qubit.lower() == "all":
        dimension = density_matrix.shape[0]
        new_density_matrix = ((1.0 - probability) * density_matrix +
                              probability * eye(dimension) / float(dimension))
        return new_density_matrix

    # For any other case, depolarize only the target qubit
    _verify_channel_inputs(density_matrix, probability, target_qubit)

    E0 = _lift_operator(sqrt(1.0 - probability) * eye(2),
                        n_qubits, target_qubit)
    E1 = _lift_operator(sqrt(probability / 3.) * array([[0.0, 1.0],
                                                        [1.0, 0.0]]),
                        n_qubits, target_qubit)
    E2 = _lift_operator(sqrt(probability / 3.) * array([[0.0, -1.0j],
                                                        [1.0j, 0.0]]),
                        n_qubits, target_qubit)
    E3 = _lift_operator(sqrt(probability / 3.) * array([[1.0, 0.0],
                                                        [0.0, -1.0]]),
                        n_qubits, target_qubit)

    new_density_matrix = (dot(E0, dot(density_matrix, E0)) +
                          dot(E1, dot(density_matrix, E1)) +
                          dot(E2, dot(density_matrix, E2)) +
                          dot(E3, dot(density_matrix, E3)))

    return new_density_matrix



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import fractions
import unittest

from openfermion.transforms import get_fermion_operator
from openfermion.utils import is_hermitian
from openfermion.utils._testing_utils import (EqualsTester, 
                                              random_interaction_operator)


class EqualsTesterTest(unittest.TestCase):

    def test_add_equality_group_correct(self):
        eq = EqualsTester(self)
    
        eq.add_equality_group(fractions.Fraction(1, 1))
    
        eq.add_equality_group(fractions.Fraction(1, 2),
                              fractions.Fraction(2, 4))
    
        eq.add_equality_group(
            fractions.Fraction(2, 3),
            fractions.Fraction(12, 18), fractions.Fraction(14, 21))
    
        eq.add_equality_group(2, 2.0, fractions.Fraction(2, 1))
    
        eq.add_equality_group([1, 2, 3], [1, 2, 3])
    
        eq.add_equality_group({'b': 3, 'a': 2}, {'a': 2, 'b': 3})
    
        eq.add_equality_group('unrelated')
    
    def test_assert_add_equality_pair(self):
        eq = EqualsTester(self)
    
        with self.assertRaises(AssertionError):
            eq.make_equality_pair(object)
    
        eq.make_equality_pair(lambda: 1)
        eq.make_equality_pair(lambda: 2)
        eq.add_equality_group(3)
    
        with self.assertRaises(AssertionError):
            eq.add_equality_group(1)
        with self.assertRaises(AssertionError):
            eq.make_equality_pair(lambda: 1)
        with self.assertRaises(AssertionError):
            eq.make_equality_pair(lambda: 3)

    def test_add_equality_group_not_equivalent(self):
        eq = EqualsTester(self)
        with self.assertRaises(AssertionError):
            eq.add_equality_group(1, 2)

    def test_add_equality_group_not_disjoint(self):
        eq = EqualsTester(self)
        eq.add_equality_group(1)
        with self.assertRaises(AssertionError):
            eq.add_equality_group(1)

    def test_add_equality_group_bad_hash(self):
        class KeyHash(object):
            def __init__(self, k, h):
                self._k = k
                self._h = h
    
            def __eq__(self, other):
                return isinstance(other, KeyHash) and self._k == other._k
    
            def __ne__(self, other):
                return not self == other
    
            def __hash__(self):
                return self._h
    
        eq = EqualsTester(self)
        eq.add_equality_group(KeyHash('a', 5), KeyHash('a', 5))
        eq.add_equality_group(KeyHash('b', 5))
        with self.assertRaises(AssertionError):
            eq.add_equality_group(KeyHash('c', 2), KeyHash('c', 3))

    def test_add_equality_group_exception_hash(self):
        class FailHash(object):
            def __hash__(self):
                raise ValueError('injected failure')
    
        eq = EqualsTester(self)
        with self.assertRaises(ValueError):
            eq.add_equality_group(FailHash())

    def test_can_fail_when_forgot_type_check(self):
        eq = EqualsTester(self)
    
        class NoTypeCheckEqualImplementation(object):
            def __init__(self):
                self.x = 1
    
            def __eq__(self, other):
                return self.x == other.x
    
            def __ne__(self, other):
                return not self == other
    
            def __hash__(self):
                return hash(self.x)
    
        with self.assertRaises(AttributeError):
            eq.add_equality_group(NoTypeCheckEqualImplementation())

    def test_fails_hash_is_default_and_inconsistent(self):
        eq = EqualsTester(self)
    
        class DefaultHashImplementation(object):
            __hash__ = object.__hash__
    
            def __init__(self):
                self.x = 1
    
            def __eq__(self, other):
                if not isinstance(other, type(self)):
                    return NotImplemented
                return self.x == other.x
    
            def __ne__(self, other):
                return not self == other
    
        with self.assertRaises(AssertionError):
            eq.make_equality_pair(DefaultHashImplementation)

    def test_fails_when_ne_is_inconsistent(self):
        eq = EqualsTester(self)
    
        class InconsistentNeImplementation(object):
            def __init__(self):
                self.x = 1
    
            def __eq__(self, other):
                if not isinstance(other, type(self)):
                    return NotImplemented
                return self.x == other.x
    
            def __ne__(self, other):
                return NotImplemented
    
            def __hash__(self):
                return hash(self.x)
    
        with self.assertRaises(AssertionError):
            eq.make_equality_pair(InconsistentNeImplementation)

    def test_fails_when_not_reflexive(self):
        eq = EqualsTester(self)
    
        class NotReflexiveImplementation(object):
            def __init__(self):
                self.x = 1
    
            def __eq__(self, other):
                if other is not self:
                    return NotImplemented
                return False
    
            def __ne__(self, other):
                return not self == other
    
        with self.assertRaises(AssertionError):
            eq.add_equality_group(NotReflexiveImplementation())
    
    def test_fails_when_not_commutative(self):
        eq = EqualsTester(self)
    
        class NotCommutativeImplementation(object):
            def __init__(self, x):
                self.x = x
    
            def __eq__(self, other):
                if not isinstance(other, type(self)):
                    return NotImplemented
                return self.x <= other.x
    
            def __ne__(self, other):
                return not self == other
    
        with self.assertRaises(AssertionError):
            eq.add_equality_group(NotCommutativeImplementation(0),
                                  NotCommutativeImplementation(1))
    
        with self.assertRaises(AssertionError):
            eq.add_equality_group(NotCommutativeImplementation(1),
                                  NotCommutativeImplementation(0))


class RandomInteractionOperatorTest(unittest.TestCase):

    def test_hermiticity(self):
        n_qubits = 5
        
        # Real case
        iop = random_interaction_operator(n_qubits, True)
        ferm_op = get_fermion_operator(iop)
        self.assertTrue(is_hermitian(ferm_op))

        # Complex case
        iop = random_interaction_operator(n_qubits, False)
        ferm_op = get_fermion_operator(iop)
        self.assertTrue(is_hermitian(ferm_op))
